"""Type inference constraints."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Final, cast
from typing_extensions import TypeGuard

import mypy.subtypes
import mypy.typeops
from mypy.argmap import ArgTypeExpander
from mypy.erasetype import erase_typevars
from mypy.maptype import map_instance_to_supertype
from mypy.nodes import (
    ARG_OPT,
    ARG_POS,
    ARG_STAR,
    ARG_STAR2,
    CONTRAVARIANT,
    COVARIANT,
    ArgKind,
    TypeInfo,
)
from mypy.types import (
    TUPLE_LIKE_INSTANCE_NAMES,
    AnyType,
    CallableType,
    DeletedType,
    ErasedType,
    Instance,
    LiteralType,
    NoneType,
    NormalizedCallableType,
    Overloaded,
    Parameters,
    ParamSpecType,
    PartialType,
    ProperType,
    TupleType,
    Type,
    TypeAliasType,
    TypedDictType,
    TypeOfAny,
    TypeQuery,
    TypeType,
    TypeVarId,
    TypeVarLikeType,
    TypeVarTupleType,
    TypeVarType,
    TypeVisitor,
    UnboundType,
    UninhabitedType,
    UnionType,
    UnpackType,
    find_unpack_in_list,
    flatten_nested_tuples,
    get_proper_type,
    has_recursive_types,
    has_type_vars,
    is_named_instance,
    split_with_prefix_and_suffix,
)
from mypy.types_utils import is_union_with_any
from mypy.typestate import type_state

if TYPE_CHECKING:
    from mypy.infer import ArgumentInferContext

SUBTYPE_OF: Final = 0
SUPERTYPE_OF: Final = 1


class Constraint:
    """A representation of a type constraint.

    It can be either T <: type or T :> type (T is a type variable).
    """

    type_var: TypeVarId
    op = 0  # SUBTYPE_OF or SUPERTYPE_OF
    target: Type

    def __init__(self, type_var: TypeVarLikeType, op: int, target: Type) -> None:
        self.type_var = type_var.id
        self.op = op
        # TODO: should we add "assert not isinstance(target, UnpackType)"?
        # UnpackType is a synthetic type, and is never valid as a constraint target.
        self.target = target
        self.origin_type_var = type_var
        # These are additional type variables that should be solved for together with type_var.
        # TODO: A cleaner solution may be to modify the return type of infer_constraints()
        # to include these instead, but this is a rather big refactoring.
        self.extra_tvars: list[TypeVarLikeType] = []

    def __repr__(self) -> str:
        op_str = "<:"
        if self.op == SUPERTYPE_OF:
            op_str = ":>"
        return f"{self.type_var} {op_str} {self.target}"

    def __hash__(self) -> int:
        return hash((self.type_var, self.op, self.target))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Constraint):
            return False
        return (self.type_var, self.op, self.target) == (other.type_var, other.op, other.target)


def infer_constraints_for_callable(
    callee: CallableType,
    arg_types: Sequence[Type | None],
    arg_kinds: list[ArgKind],
    arg_names: Sequence[str | None] | None,
    formal_to_actual: list[list[int]],
    context: ArgumentInferContext,
) -> list[Constraint]:
    """Infer type variable constraints for a callable and actual arguments.

    Return a list of constraints.
    """
    constraints: list[Constraint] = []
    mapper = ArgTypeExpander(context)

    param_spec = callee.param_spec()
    param_spec_arg_types = []
    param_spec_arg_names = []
    param_spec_arg_kinds = []

    incomplete_star_mapping = False
    for i, actuals in enumerate(formal_to_actual):  # TODO: isn't this `enumerate(arg_types)`?
        for actual in actuals:
            if actual is None and callee.arg_kinds[i] in (ARG_STAR, ARG_STAR2):  # type: ignore[unreachable]
                # We can't use arguments to infer ParamSpec constraint, if only some
                # are present in the current inference pass.
                incomplete_star_mapping = True  # type: ignore[unreachable]
                break

    for i, actuals in enumerate(formal_to_actual):
        if isinstance(callee.arg_types[i], UnpackType):
            unpack_type = callee.arg_types[i]
            assert isinstance(unpack_type, UnpackType)

            # In this case we are binding all the actuals to *args,
            # and we want a constraint that the typevar tuple being unpacked
            # is equal to a type list of all the actuals.
            actual_types = []

            unpacked_type = get_proper_type(unpack_type.type)
            if isinstance(unpacked_type, TypeVarTupleType):
                tuple_instance = unpacked_type.tuple_fallback
            elif isinstance(unpacked_type, TupleType):
                tuple_instance = unpacked_type.partial_fallback
            else:
                assert False, "mypy bug: unhandled constraint inference case"

            for actual in actuals:
                actual_arg_type = arg_types[actual]
                if actual_arg_type is None:
                    continue

                expanded_actual = mapper.expand_actual_type(
                    actual_arg_type,
                    arg_kinds[actual],
                    callee.arg_names[i],
                    callee.arg_kinds[i],
                    allow_unpack=True,
                )

                if arg_kinds[actual] != ARG_STAR or isinstance(
                    get_proper_type(actual_arg_type), TupleType
                ):
                    actual_types.append(expanded_actual)
                else:
                    # If we are expanding an iterable inside * actual, append a homogeneous item instead
                    actual_types.append(
                        UnpackType(tuple_instance.copy_modified(args=[expanded_actual]))
                    )

            if isinstance(unpacked_type, TypeVarTupleType):
                constraints.append(
                    Constraint(
                        unpacked_type,
                        SUPERTYPE_OF,
                        TupleType(actual_types, unpacked_type.tuple_fallback),
                    )
                )
            elif isinstance(unpacked_type, TupleType):
                # Prefixes get converted to positional args, so technically the only case we
                # should have here is like Tuple[Unpack[Ts], Y1, Y2, Y3]. If this turns out
                # not to hold we can always handle the prefixes too.
                inner_unpack = unpacked_type.items[0]
                assert isinstance(inner_unpack, UnpackType)
                inner_unpacked_type = get_proper_type(inner_unpack.type)
                suffix_len = len(unpacked_type.items) - 1
                if isinstance(inner_unpacked_type, TypeVarTupleType):
                    # Variadic item can be either *Ts...
                    constraints.append(
                        Constraint(
                            inner_unpacked_type,
                            SUPERTYPE_OF,
                            TupleType(
                                actual_types[:-suffix_len], inner_unpacked_type.tuple_fallback
                            ),
                        )
                    )
                else:
                    # ...or it can be a homogeneous tuple.
                    assert (
                        isinstance(inner_unpacked_type, Instance)
                        and inner_unpacked_type.type.fullname == "builtins.tuple"
                    )
                    for at in actual_types[:-suffix_len]:
                        constraints.extend(
                            infer_constraints(inner_unpacked_type.args[0], at, SUPERTYPE_OF)
                        )
                # Now handle the suffix (if any).
                if suffix_len:
                    for tt, at in zip(unpacked_type.items[1:], actual_types[-suffix_len:]):
                        constraints.extend(infer_constraints(tt, at, SUPERTYPE_OF))
            else:
                assert False, "mypy bug: unhandled constraint inference case"
        else:
            for actual in actuals:
                actual_arg_type = arg_types[actual]
                if actual_arg_type is None:
                    continue

                if param_spec and callee.arg_kinds[i] in (ARG_STAR, ARG_STAR2):
                    # If actual arguments are mapped to ParamSpec type, we can't infer individual
                    # constraints, instead store them and infer single constraint at the end.
                    # It is impossible to map actual kind to formal kind, so use some heuristic.
                    # This inference is used as a fallback, so relying on heuristic should be OK.
                    if not incomplete_star_mapping:
                        param_spec_arg_types.append(
                            mapper.expand_actual_type(
                                actual_arg_type, arg_kinds[actual], None, arg_kinds[actual]
                            )
                        )
                        actual_kind = arg_kinds[actual]
                        param_spec_arg_kinds.append(
                            ARG_POS if actual_kind not in (ARG_STAR, ARG_STAR2) else actual_kind
                        )
                        param_spec_arg_names.append(arg_names[actual] if arg_names else None)
                else:
                    actual_type = mapper.expand_actual_type(
                        actual_arg_type,
                        arg_kinds[actual],
                        callee.arg_names[i],
                        callee.arg_kinds[i],
                    )
                    c = infer_constraints(callee.arg_types[i], actual_type, SUPERTYPE_OF)
                    constraints.extend(c)
    if (
        param_spec
        and not any(c.type_var == param_spec.id for c in constraints)
        and not incomplete_star_mapping
    ):
        # Use ParamSpec constraint from arguments only if there are no other constraints,
        # since as explained above it is quite ad-hoc.
        constraints.append(
            Constraint(
                param_spec,
                SUPERTYPE_OF,
                Parameters(
                    arg_types=param_spec_arg_types,
                    arg_kinds=param_spec_arg_kinds,
                    arg_names=param_spec_arg_names,
                    imprecise_arg_kinds=True,
                ),
            )
        )
    if any(isinstance(v, ParamSpecType) for v in callee.variables):
        # As a perf optimization filter imprecise constraints only when we can have them.
        constraints = filter_imprecise_kinds(constraints)
    return constraints


def infer_constraints(
    template: Type, actual: Type, direction: int, skip_neg_op: bool = False
) -> list[Constraint]:
    """Infer type constraints.

    Match a template type, which may contain type variable references,
    recursively against a type which does not contain (the same) type
    variable references. The result is a list of type constrains of
    form 'T is a supertype/subtype of x', where T is a type variable
    present in the template and x is a type without reference to type
    variables present in the template.

    Assume T and S are type variables. Now the following results can be
    calculated (read as '(template, actual) --> result'):

      (T, X)            -->  T :> X
      (X[T], X[Y])      -->  T <: Y and T :> Y
      ((T, T), (X, Y))  -->  T :> X and T :> Y
      ((T, S), (X, Y))  -->  T :> X and S :> Y
      (X[T], Any)       -->  T <: Any and T :> Any

    The constraints are represented as Constraint objects. If skip_neg_op == True,
    then skip adding reverse (polymorphic) constraints (since this is already a call
    to infer such constraints).
    """
    if any(
        get_proper_type(template) == get_proper_type(t)
        and get_proper_type(actual) == get_proper_type(a)
        for (t, a) in reversed(type_state.inferring)
    ):
        return []
    if has_recursive_types(template) or isinstance(get_proper_type(template), Instance):
        # This case requires special care because it may cause infinite recursion.
        # Note that we include Instances because the may be recursive as str(Sequence[str]).
        if not has_type_vars(template):
            # Return early on an empty branch.
            return []
        type_state.inferring.append((template, actual))
        res = _infer_constraints(template, actual, direction, skip_neg_op)
        type_state.inferring.pop()
        return res
    return _infer_constraints(template, actual, direction, skip_neg_op)


def _infer_constraints(
    template: Type, actual: Type, direction: int, skip_neg_op: bool
) -> list[Constraint]:
    orig_template = template
    template = get_proper_type(template)
    actual = get_proper_type(actual)

    # Type inference shouldn't be affected by whether union types have been simplified.
    # We however keep any ErasedType items, so that the caller will see it when using
    # checkexpr.has_erased_component().
    if isinstance(template, UnionType):
        template = mypy.typeops.make_simplified_union(template.items, keep_erased=True)
    if isinstance(actual, UnionType):
        actual = mypy.typeops.make_simplified_union(actual.items, keep_erased=True)

    # Ignore Any types from the type suggestion engine to avoid them
    # causing us to infer Any in situations where a better job could
    # be done otherwise. (This can produce false positives but that
    # doesn't really matter because it is all heuristic anyway.)
    if isinstance(actual, AnyType) and actual.type_of_any == TypeOfAny.suggestion_engine:
        return []

    # type[A | B] is always represented as type[A] | type[B] internally.
    # This makes our constraint solver choke on type[T] <: type[A] | type[B],
    # solving T as generic meet(A, B) which is often `object`. Force unwrap such unions
    # if both sides are type[...] or unions thereof. See `testTypeVarType` test
    type_type_unwrapped = False
    if _is_type_type(template) and _is_type_type(actual):
        type_type_unwrapped = True
        template = _unwrap_type_type(template)
        actual = _unwrap_type_type(actual)

    # If the template is simply a type variable, emit a Constraint directly.
    # We need to handle this case before handling Unions for two reasons:
    #  1. "T <: Union[U1, U2]" is not equivalent to "T <: U1 or T <: U2",
    #     because T can itself be a union (notably, Union[U1, U2] itself).
    #  2. "T :> Union[U1, U2]" is logically equivalent to "T :> U1 and
    #     T :> U2", but they are not equivalent to the constraint solver,
    #     which never introduces new Union types (it uses join() instead).
    if isinstance(template, TypeVarType):
        return [Constraint(template, direction, actual)]

    if (
        isinstance(actual, TypeVarType)
        and not actual.id.is_meta_var()
        and direction == SUPERTYPE_OF
    ):
        # Unless template is also a type variable (or a union that contains one), using the upper
        # bound for inference will usually give better result for actual that is a type variable.
        if not isinstance(template, UnionType) or not any(
            isinstance(t, TypeVarType) for t in template.items
        ):
            actual = get_proper_type(actual.upper_bound)

    # Now handle the case of either template or actual being a Union.
    # For a Union to be a subtype of another type, every item of the Union
    # must be a subtype of that type, so concatenate the constraints.
    if direction == SUBTYPE_OF and isinstance(template, UnionType):
        res = []
        for t_item in template.items:
            res.extend(infer_constraints(t_item, actual, direction))
        return res
    if direction == SUPERTYPE_OF and isinstance(actual, UnionType):
        res = []
        for a_item in actual.items:
            # `orig_template` has to be preserved intact in case it's recursive.
            # If we unwrapped ``type[...]`` previously, wrap the item back again,
            # as ``type[...]`` can't be removed from `orig_template`.
            if type_type_unwrapped:
                a_item = TypeType.make_normalized(a_item)
            res.extend(infer_constraints(orig_template, a_item, direction))
        return res

    # Now the potential subtype is known not to be a Union or a type
    # variable that we are solving for. In that case, for a Union to
    # be a supertype of the potential subtype, some item of the Union
    # must be a supertype of it.
    if direction == SUBTYPE_OF and isinstance(actual, UnionType):
        # If some of items is not a complete type, disregard that.
        items = simplify_away_incomplete_types(actual.items)
        # We infer constraints eagerly -- try to find constraints for a type
        # variable if possible. This seems to help with some real-world
        # use cases.
        return any_constraints(
            [infer_constraints_if_possible(template, a_item, direction) for a_item in items],
            eager=True,
        )
    if direction == SUPERTYPE_OF and isinstance(template, UnionType):
        # When the template is a union, we are okay with leaving some
        # type variables indeterminate. This helps with some special
        # cases, though this isn't very principled.
        result = any_constraints(
            [
                infer_constraints_if_possible(t_item, actual, direction)
                for t_item in template.items
            ],
            eager=isinstance(actual, AnyType),
        )
        if result:
            return result
        elif has_recursive_types(template) and not has_recursive_types(actual):
            return handle_recursive_union(template, actual, direction)
        return []

    # Remaining cases are handled by ConstraintBuilderVisitor.
    return template.accept(ConstraintBuilderVisitor(actual, direction, skip_neg_op))


def _is_type_type(tp: ProperType) -> TypeGuard[TypeType | UnionType]:
    """Is ``tp`` a ``type[...]`` or a union thereof?

    ``Type[A | B]`` is internally represented as ``type[A] | type[B]``, and this
    troubles the solver sometimes.
    """
    return (
        isinstance(tp, TypeType)
        or isinstance(tp, UnionType)
        and all(isinstance(get_proper_type(o), TypeType) for o in tp.items)
    )


def _unwrap_type_type(tp: TypeType | UnionType) -> ProperType:
    """Extract the inner type from ``type[...]`` expression or a union thereof."""
    if isinstance(tp, TypeType):
        return tp.item
    return UnionType.make_union([cast(TypeType, get_proper_type(o)).item for o in tp.items])


def infer_constraints_if_possible(
    template: Type, actual: Type, direction: int
) -> list[Constraint] | None:
    """Like infer_constraints, but return None if the input relation is
    known to be unsatisfiable, for example if template=List[T] and actual=int.
    (In this case infer_constraints would return [], just like it would for
    an automatically satisfied relation like template=List[T] and actual=object.)
    """
    if direction == SUBTYPE_OF and not mypy.subtypes.is_subtype(erase_typevars(template), actual):
        return None
    if direction == SUPERTYPE_OF and not mypy.subtypes.is_subtype(
        actual, erase_typevars(template)
    ):
        return None
    if (
        direction == SUPERTYPE_OF
        and isinstance(template, TypeVarType)
        and not mypy.subtypes.is_subtype(actual, erase_typevars(template.upper_bound))
    ):
        # This is not caught by the above branch because of the erase_typevars() call,
        # that would return 'Any' for a type variable.
        return None
    return infer_constraints(template, actual, direction)


def select_trivial(options: Sequence[list[Constraint] | None]) -> list[list[Constraint]]:
    """Select only those lists where each item is a constraint against Any."""
    res = []
    for option in options:
        if option is None:
            continue
        if all(isinstance(get_proper_type(c.target), AnyType) for c in option):
            res.append(option)
    return res


def merge_with_any(constraint: Constraint) -> Constraint:
    """Transform a constraint target into a union with given Any type."""
    target = constraint.target
    if is_union_with_any(target):
        # Do not produce redundant unions.
        return constraint
    # TODO: if we will support multiple sources Any, use this here instead.
    any_type = AnyType(TypeOfAny.implementation_artifact)
    return Constraint(
        constraint.origin_type_var,
        constraint.op,
        UnionType.make_union([target, any_type], target.line, target.column),
    )


def handle_recursive_union(template: UnionType, actual: Type, direction: int) -> list[Constraint]:
    # This is a hack to special-case things like Union[T, Inst[T]] in recursive types. Although
    # it is quite arbitrary, it is a relatively common pattern, so we should handle it well.
    # This function may be called when inferring against such union resulted in different
    # constraints for each item. Normally we give up in such case, but here we instead split
    # the union in two parts, and try inferring sequentially.
    non_type_var_items = [t for t in template.items if not isinstance(t, TypeVarType)]
    type_var_items = [t for t in template.items if isinstance(t, TypeVarType)]
    return infer_constraints(
        UnionType.make_union(non_type_var_items), actual, direction
    ) or infer_constraints(UnionType.make_union(type_var_items), actual, direction)


def any_constraints(options: list[list[Constraint] | None], *, eager: bool) -> list[Constraint]:
    """Deduce what we can from a collection of constraint lists.

    It's a given that at least one of the lists must be satisfied. A
    None element in the list of options represents an unsatisfiable
    constraint and is ignored.  Ignore empty constraint lists if eager
    is true -- they are always trivially satisfiable.
    """
    if eager:
        valid_options = [option for option in options if option]
    else:
        valid_options = [option for option in options if option is not None]

    if not valid_options:
        return []

    if len(valid_options) == 1:
        return valid_options[0]

    if all(is_same_constraints(valid_options[0], c) for c in valid_options[1:]):
        # Multiple sets of constraints that are all the same. Just pick any one of them.
        return valid_options[0]

    if all(is_similar_constraints(valid_options[0], c) for c in valid_options[1:]):
        # All options have same structure. In this case we can merge-in trivial
        # options (i.e. those that only have Any) and try again.
        # TODO: More generally, if a given (variable, direction) pair appears in
        # every option, combine the bounds with meet/join always, not just for Any.
        trivial_options = select_trivial(valid_options)
        if trivial_options and len(trivial_options) < len(valid_options):
            merged_options = []
            for option in valid_options:
                if option in trivial_options:
                    continue
                merged_options.append([merge_with_any(c) for c in option])
            return any_constraints(list(merged_options), eager=eager)

    # If normal logic didn't work, try excluding trivially unsatisfiable constraint (due to
    # upper bounds) from each option, and comparing them again.
    filtered_options = [filter_satisfiable(o) for o in options]
    if filtered_options != options:
        return any_constraints(filtered_options, eager=eager)

    # Try harder: if that didn't work, try to strip typevars that aren't meta vars.
    # Note this is what we would always do, but unfortunately some callers may not
    # set the meta var status correctly (for historical reasons), so we use this as
    # a fallback only.
    filtered_options = [exclude_non_meta_vars(o) for o in options]
    if filtered_options != options:
        return any_constraints(filtered_options, eager=eager)

    # Otherwise, there are either no valid options or multiple, inconsistent valid
    # options. Give up and deduce nothing.
    return []


def filter_satisfiable(option: list[Constraint] | None) -> list[Constraint] | None:
    """Keep only constraints that can possibly be satisfied.

    Currently, we filter out constraints where target is not a subtype of the upper bound.
    Since those can be never satisfied. We may add more cases in future if it improves type
    inference.
    """
    if not option:
        return option

    satisfiable = []
    for c in option:
        if isinstance(c.origin_type_var, TypeVarType) and c.origin_type_var.values:
            if any(
                mypy.subtypes.is_subtype(c.target, value) for value in c.origin_type_var.values
            ):
                satisfiable.append(c)
        elif mypy.subtypes.is_subtype(c.target, c.origin_type_var.upper_bound):
            satisfiable.append(c)
    if not satisfiable:
        return None
    return satisfiable


def exclude_non_meta_vars(option: list[Constraint] | None) -> list[Constraint] | None:
    # If we had an empty list, keep it intact
    if not option:
        return option
    # However, if none of the options actually references meta vars, better remove
    # this constraint entirely.
    return [c for c in option if c.type_var.is_meta_var()] or None


def is_same_constraints(x: list[Constraint], y: list[Constraint]) -> bool:
    for c1 in x:
        if not any(is_same_constraint(c1, c2) for c2 in y):
            return False
    for c1 in y:
        if not any(is_same_constraint(c1, c2) for c2 in x):
            return False
    return True


def is_same_constraint(c1: Constraint, c2: Constraint) -> bool:
    # Ignore direction when comparing constraints against Any.
    skip_op_check = isinstance(get_proper_type(c1.target), AnyType) and isinstance(
        get_proper_type(c2.target), AnyType
    )
    return (
        c1.type_var == c2.type_var
        and (c1.op == c2.op or skip_op_check)
        and mypy.subtypes.is_same_type(c1.target, c2.target)
    )


def is_similar_constraints(x: list[Constraint], y: list[Constraint]) -> bool:
    """Check that two lists of constraints have similar structure.

    This means that each list has same type variable plus direction pairs (i.e we
    ignore the target). Except for constraints where target is Any type, there
    we ignore direction as well.
    """
    return _is_similar_constraints(x, y) and _is_similar_constraints(y, x)


def _is_similar_constraints(x: list[Constraint], y: list[Constraint]) -> bool:
    """Check that every constraint in the first list has a similar one in the second.

    See docstring above for definition of similarity.
    """
    for c1 in x:
        has_similar = False
        for c2 in y:
            # Ignore direction when either constraint is against Any.
            skip_op_check = isinstance(get_proper_type(c1.target), AnyType) or isinstance(
                get_proper_type(c2.target), AnyType
            )
            if c1.type_var == c2.type_var and (c1.op == c2.op or skip_op_check):
                has_similar = True
                break
        if not has_similar:
            return False
    return True


def simplify_away_incomplete_types(types: Iterable[Type]) -> list[Type]:
    complete = [typ for typ in types if is_complete_type(typ)]
    if complete:
        return complete
    else:
        return list(types)


def is_complete_type(typ: Type) -> bool:
    """Is a type complete?

    A complete doesn't have uninhabited type components or (when not in strict
    optional mode) None components.
    """
    return typ.accept(CompleteTypeVisitor())


class CompleteTypeVisitor(TypeQuery[bool]):
    def __init__(self) -> None:
        super().__init__(all)

    def visit_uninhabited_type(self, t: UninhabitedType) -> bool:
        return False


class ConstraintBuilderVisitor(TypeVisitor[list[Constraint]]):
    """Visitor class for inferring type constraints."""

    # The type that is compared against a template
    # TODO: The value may be None. Is that actually correct?
    actual: ProperType

    def __init__(self, actual: ProperType, direction: int, skip_neg_op: bool) -> None:
        # Direction must be SUBTYPE_OF or SUPERTYPE_OF.
        self.actual = actual
        self.direction = direction
        # Whether to skip polymorphic inference (involves inference in opposite direction)
        # this is used to prevent infinite recursion when both template and actual are
        # generic callables.
        self.skip_neg_op = skip_neg_op

    # Trivial leaf types

    def visit_unbound_type(self, template: UnboundType) -> list[Constraint]:
        return []

    def visit_any(self, template: AnyType) -> list[Constraint]:
        return []

    def visit_none_type(self, template: NoneType) -> list[Constraint]:
        return []

    def visit_uninhabited_type(self, template: UninhabitedType) -> list[Constraint]:
        return []

    def visit_erased_type(self, template: ErasedType) -> list[Constraint]:
        return []

    def visit_deleted_type(self, template: DeletedType) -> list[Constraint]:
        return []

    def visit_literal_type(self, template: LiteralType) -> list[Constraint]:
        return []

    # Errors

    def visit_partial_type(self, template: PartialType) -> list[Constraint]:
        # We can't do anything useful with a partial type here.
        assert False, "Internal error"

    # Non-trivial leaf type

    def visit_type_var(self, template: TypeVarType) -> list[Constraint]:
        assert False, (
            "Unexpected TypeVarType in ConstraintBuilderVisitor"
            " (should have been handled in infer_constraints)"
        )

    def visit_param_spec(self, template: ParamSpecType) -> list[Constraint]:
        # Can't infer ParamSpecs from component values (only via Callable[P, T]).
        return []

    def visit_type_var_tuple(self, template: TypeVarTupleType) -> list[Constraint]:
        raise NotImplementedError

    def visit_unpack_type(self, template: UnpackType) -> list[Constraint]:
        raise RuntimeError("Mypy bug: unpack should be handled at a higher level.")

    def visit_parameters(self, template: Parameters) -> list[Constraint]:
        # Constraining Any against C[P] turns into infer_against_any([P], Any)
        if isinstance(self.actual, AnyType):
            return self.infer_against_any(template.arg_types, self.actual)
        if type_state.infer_polymorphic and isinstance(self.actual, Parameters):
            # For polymorphic inference we need to be able to infer secondary constraints
            # in situations like [x: T] <: P <: [x: int].
            return infer_callable_arguments_constraints(template, self.actual, self.direction)
        if type_state.infer_polymorphic and isinstance(self.actual, ParamSpecType):
            # Similar for [x: T] <: Q <: Concatenate[int, P].
            return infer_callable_arguments_constraints(
                template, self.actual.prefix, self.direction
            )
        # There also may be unpatched types after a user error, simply ignore them.
        return []

    # Non-leaf types

    def visit_instance(self, template: Instance) -> list[Constraint]:
        original_actual = actual = self.actual
        res: list[Constraint] = []
        if isinstance(actual, (CallableType, Overloaded)) and template.type.is_protocol:
            if "__call__" in template.type.protocol_members:
                # Special case: a generic callback protocol
                if not any(template == t for t in template.type.inferring):
                    template.type.inferring.append(template)
                    call = mypy.subtypes.find_member(
                        "__call__", template, actual, is_operator=True
                    )
                    assert call is not None
                    if (
                        self.direction == SUPERTYPE_OF
                        and mypy.subtypes.is_subtype(actual, erase_typevars(call))
                        or self.direction == SUBTYPE_OF
                        and mypy.subtypes.is_subtype(erase_typevars(call), actual)
                    ):
                        res.extend(infer_constraints(call, actual, self.direction))
                    template.type.inferring.pop()
        if isinstance(actual, CallableType) and actual.fallback is not None:
            if (
                actual.is_type_obj()
                and template.type.is_protocol
                and self.direction == SUPERTYPE_OF
            ):
                ret_type = get_proper_type(actual.ret_type)
                if isinstance(ret_type, TupleType):
                    ret_type = mypy.typeops.tuple_fallback(ret_type)
                if isinstance(ret_type, Instance):
                    res.extend(
                        self.infer_constraints_from_protocol_members(
                            ret_type, template, ret_type, template, class_obj=True
                        )
                    )
            actual = actual.fallback
        if isinstance(actual, TypeType) and template.type.is_protocol:
            if self.direction == SUPERTYPE_OF:
                a_item = actual.item
                if isinstance(a_item, Instance):
                    res.extend(
                        self.infer_constraints_from_protocol_members(
                            a_item, template, a_item, template, class_obj=True
                        )
                    )
                # Infer constraints for Type[T] via metaclass of T when it makes sense.
                if isinstance(a_item, TypeVarType):
                    a_item = get_proper_type(a_item.upper_bound)
                if isinstance(a_item, Instance) and a_item.type.metaclass_type:
                    res.extend(
                        self.infer_constraints_from_protocol_members(
                            a_item.type.metaclass_type, template, actual, template
                        )
                    )

        if isinstance(actual, Overloaded) and actual.fallback is not None:
            actual = actual.fallback
        if isinstance(actual, TypedDictType):
            actual = actual.as_anonymous().fallback
        if isinstance(actual, LiteralType):
            actual = actual.fallback
        if isinstance(actual, Instance):
            instance = actual
            erased = erase_typevars(template)
            assert isinstance(erased, Instance)  # type: ignore[misc]
            # We always try nominal inference if possible,
            # it is much faster than the structural one.
            if self.direction == SUBTYPE_OF and template.type.has_base(instance.type.fullname):
                mapped = map_instance_to_supertype(template, instance.type)
                tvars = mapped.type.defn.type_vars

                if instance.type.has_type_var_tuple_type:
                    # Variadic types need special handling to map each type argument to
                    # the correct corresponding type variable.
                    assert instance.type.type_var_tuple_prefix is not None
                    assert instance.type.type_var_tuple_suffix is not None
                    prefix_len = instance.type.type_var_tuple_prefix
                    suffix_len = instance.type.type_var_tuple_suffix
                    tvt = instance.type.defn.type_vars[prefix_len]
                    assert isinstance(tvt, TypeVarTupleType)
                    fallback = tvt.tuple_fallback
                    i_prefix, i_middle, i_suffix = split_with_prefix_and_suffix(
                        instance.args, prefix_len, suffix_len
                    )
                    m_prefix, m_middle, m_suffix = split_with_prefix_and_suffix(
                        mapped.args, prefix_len, suffix_len
                    )
                    instance_args = i_prefix + (TupleType(list(i_middle), fallback),) + i_suffix
                    mapped_args = m_prefix + (TupleType(list(m_middle), fallback),) + m_suffix
                else:
                    mapped_args = mapped.args
                    instance_args = instance.args

                # N.B: We use zip instead of indexing because the lengths might have
                # mismatches during daemon reprocessing.
                for tvar, mapped_arg, instance_arg in zip(tvars, mapped_args, instance_args):
                    if isinstance(tvar, TypeVarType):
                        # The constraints for generic type parameters depend on variance.
                        # Include constraints from both directions if invariant.
                        if tvar.variance != CONTRAVARIANT:
                            res.extend(infer_constraints(mapped_arg, instance_arg, self.direction))
                        if tvar.variance != COVARIANT:
                            res.extend(
                                infer_constraints(mapped_arg, instance_arg, neg_op(self.direction))
                            )
                    elif isinstance(tvar, ParamSpecType) and isinstance(mapped_arg, ParamSpecType):
                        prefix = mapped_arg.prefix
                        if isinstance(instance_arg, Parameters):
                            # No such thing as variance for ParamSpecs, consider them invariant
                            # TODO: constraints between prefixes using
                            # infer_callable_arguments_constraints()
                            suffix: Type = instance_arg.copy_modified(
                                instance_arg.arg_types[len(prefix.arg_types) :],
                                instance_arg.arg_kinds[len(prefix.arg_kinds) :],
                                instance_arg.arg_names[len(prefix.arg_names) :],
                            )
                            res.append(Constraint(mapped_arg, SUBTYPE_OF, suffix))
                            res.append(Constraint(mapped_arg, SUPERTYPE_OF, suffix))
                        elif isinstance(instance_arg, ParamSpecType):
                            suffix = instance_arg.copy_modified(
                                prefix=Parameters(
                                    instance_arg.prefix.arg_types[len(prefix.arg_types) :],
                                    instance_arg.prefix.arg_kinds[len(prefix.arg_kinds) :],
                                    instance_arg.prefix.arg_names[len(prefix.arg_names) :],
                                )
                            )
                            res.append(Constraint(mapped_arg, SUBTYPE_OF, suffix))
                            res.append(Constraint(mapped_arg, SUPERTYPE_OF, suffix))
                    elif isinstance(tvar, TypeVarTupleType):
                        # Handle variadic type variables covariantly for consistency.
                        res.extend(infer_constraints(mapped_arg, instance_arg, self.direction))

                return res
            elif self.direction == SUPERTYPE_OF and instance.type.has_base(template.type.fullname):
                mapped = map_instance_to_supertype(instance, template.type)
                tvars = template.type.defn.type_vars
                if template.type.has_type_var_tuple_type:
                    # Variadic types need special handling to map each type argument to
                    # the correct corresponding type variable.
                    assert template.type.type_var_tuple_prefix is not None
                    assert template.type.type_var_tuple_suffix is not None
                    prefix_len = template.type.type_var_tuple_prefix
                    suffix_len = template.type.type_var_tuple_suffix
                    tvt = template.type.defn.type_vars[prefix_len]
                    assert isinstance(tvt, TypeVarTupleType)
                    fallback = tvt.tuple_fallback
                    t_prefix, t_middle, t_suffix = split_with_prefix_and_suffix(
                        template.args, prefix_len, suffix_len
                    )
                    m_prefix, m_middle, m_suffix = split_with_prefix_and_suffix(
                        mapped.args, prefix_len, suffix_len
                    )
                    template_args = t_prefix + (TupleType(list(t_middle), fallback),) + t_suffix
                    mapped_args = m_prefix + (TupleType(list(m_middle), fallback),) + m_suffix
                else:
                    mapped_args = mapped.args
                    template_args = template.args
                # N.B: We use zip instead of indexing because the lengths might have
                # mismatches during daemon reprocessing.
                for tvar, mapped_arg, template_arg in zip(tvars, mapped_args, template_args):
                    if isinstance(tvar, TypeVarType):
                        # The constraints for generic type parameters depend on variance.
                        # Include constraints from both directions if invariant.
                        if tvar.variance != CONTRAVARIANT:
                            res.extend(infer_constraints(template_arg, mapped_arg, self.direction))
                        if tvar.variance != COVARIANT:
                            res.extend(
                                infer_constraints(template_arg, mapped_arg, neg_op(self.direction))
                            )
                    elif isinstance(tvar, ParamSpecType) and isinstance(
                        template_arg, ParamSpecType
                    ):
                        prefix = template_arg.prefix
                        if isinstance(mapped_arg, Parameters):
                            # No such thing as variance for ParamSpecs, consider them invariant
                            # TODO: constraints between prefixes using
                            # infer_callable_arguments_constraints()
                            suffix = mapped_arg.copy_modified(
                                mapped_arg.arg_types[len(prefix.arg_types) :],
                                mapped_arg.arg_kinds[len(prefix.arg_kinds) :],
                                mapped_arg.arg_names[len(prefix.arg_names) :],
                            )
                            res.append(Constraint(template_arg, SUBTYPE_OF, suffix))
                            res.append(Constraint(template_arg, SUPERTYPE_OF, suffix))
                        elif isinstance(mapped_arg, ParamSpecType):
                            suffix = mapped_arg.copy_modified(
                                prefix=Parameters(
                                    mapped_arg.prefix.arg_types[len(prefix.arg_types) :],
                                    mapped_arg.prefix.arg_kinds[len(prefix.arg_kinds) :],
                                    mapped_arg.prefix.arg_names[len(prefix.arg_names) :],
                                )
                            )
                            res.append(Constraint(template_arg, SUBTYPE_OF, suffix))
                            res.append(Constraint(template_arg, SUPERTYPE_OF, suffix))
                    elif isinstance(tvar, TypeVarTupleType):
                        # Consider variadic type variables to be invariant.
                        res.extend(infer_constraints(template_arg, mapped_arg, SUBTYPE_OF))
                        res.extend(infer_constraints(template_arg, mapped_arg, SUPERTYPE_OF))
                return res
            if (
                template.type.is_protocol
                and self.direction == SUPERTYPE_OF
                and
                # We avoid infinite recursion for structural subtypes by checking
                # whether this type already appeared in the inference chain.
                # This is a conservative way to break the inference cycles.
                # It never produces any "false" constraints but gives up soon
                # on purely structural inference cycles, see #3829.
                # Note that we use is_protocol_implementation instead of is_subtype
                # because some type may be considered a subtype of a protocol
                # due to _promote, but still not implement the protocol.
                not any(template == t for t in reversed(template.type.inferring))
                and mypy.subtypes.is_protocol_implementation(instance, erased, skip=["__call__"])
            ):
                template.type.inferring.append(template)
                res.extend(
                    self.infer_constraints_from_protocol_members(
                        instance, template, original_actual, template
                    )
                )
                template.type.inferring.pop()
                return res
            elif (
                instance.type.is_protocol
                and self.direction == SUBTYPE_OF
                and
                # We avoid infinite recursion for structural subtypes also here.
                not any(instance == i for i in reversed(instance.type.inferring))
                and mypy.subtypes.is_protocol_implementation(erased, instance, skip=["__call__"])
            ):
                instance.type.inferring.append(instance)
                res.extend(
                    self.infer_constraints_from_protocol_members(
                        instance, template, template, instance
                    )
                )
                instance.type.inferring.pop()
                return res
        if res:
            return res

        if isinstance(actual, AnyType):
            return self.infer_against_any(template.args, actual)
        if (
            isinstance(actual, TupleType)
            and is_named_instance(template, TUPLE_LIKE_INSTANCE_NAMES)
            and self.direction == SUPERTYPE_OF
        ):
            for item in actual.items:
                if isinstance(item, UnpackType):
                    unpacked = get_proper_type(item.type)
                    if isinstance(unpacked, TypeVarTupleType):
                        # Cannot infer anything for T from [T, ...] <: *Ts
                        continue
                    assert (
                        isinstance(unpacked, Instance)
                        and unpacked.type.fullname == "builtins.tuple"
                    )
                    item = unpacked.args[0]
                cb = infer_constraints(template.args[0], item, SUPERTYPE_OF)
                res.extend(cb)
            return res
        elif isinstance(actual, TupleType) and self.direction == SUPERTYPE_OF:
            return infer_constraints(template, mypy.typeops.tuple_fallback(actual), self.direction)
        elif isinstance(actual, TypeVarType):
            if not actual.values and not actual.id.is_meta_var():
                return infer_constraints(template, actual.upper_bound, self.direction)
            return []
        elif isinstance(actual, ParamSpecType):
            return infer_constraints(template, actual.upper_bound, self.direction)
        elif isinstance(actual, TypeVarTupleType):
            raise NotImplementedError
        else:
            return []

    def infer_constraints_from_protocol_members(
        self,
        instance: Instance,
        template: Instance,
        subtype: Type,
        protocol: Instance,
        class_obj: bool = False,
    ) -> list[Constraint]:
        """Infer constraints for situations where either 'template' or 'instance' is a protocol.

        The 'protocol' is the one of two that is an instance of protocol type, 'subtype'
        is the type used to bind self during inference. Currently, we just infer constrains for
        every protocol member type (both ways for settable members).
        """
        res = []
        for member in protocol.type.protocol_members:
            inst = mypy.subtypes.find_member(member, instance, subtype, class_obj=class_obj)
            temp = mypy.subtypes.find_member(member, template, subtype)
            if inst is None or temp is None:
                if member == "__call__":
                    continue
                return []  # See #11020
            # The above is safe since at this point we know that 'instance' is a subtype
            # of (erased) 'template', therefore it defines all protocol members
            if class_obj:
                # For class objects we must only infer constraints if possible, otherwise it
                # can lead to confusion between class and instance, for example StrEnum is
                # Iterable[str] for an instance, but Iterable[StrEnum] for a class object.
                if not mypy.subtypes.is_subtype(
                    inst, erase_typevars(temp), ignore_pos_arg_names=True
                ):
                    continue
            # This exception matches the one in typeops.py, see PR #14121 for context.
            if member == "__call__" and instance.type.is_metaclass(precise=True):
                continue
            res.extend(infer_constraints(temp, inst, self.direction))
            if mypy.subtypes.IS_SETTABLE in mypy.subtypes.get_member_flags(member, protocol):
                # Settable members are invariant, add opposite constraints
                res.extend(infer_constraints(temp, inst, neg_op(self.direction)))
        return res

    def visit_callable_type(self, template: CallableType) -> list[Constraint]:
        # Normalize callables before matching against each other.
        # Note that non-normalized callables can be created in annotations
        # using e.g. callback protocols.
        # TODO: check that callables match? Ideally we should not infer constraints
        # callables that can never be subtypes of one another in given direction.
        template = template.with_unpacked_kwargs().with_normalized_var_args()
        extra_tvars = False
        if isinstance(self.actual, CallableType):
            res: list[Constraint] = []
            cactual = self.actual.with_unpacked_kwargs().with_normalized_var_args()
            param_spec = template.param_spec()

            template_ret_type, cactual_ret_type = template.ret_type, cactual.ret_type
            if template.type_guard is not None and cactual.type_guard is not None:
                template_ret_type = template.type_guard
                cactual_ret_type = cactual.type_guard

            if template.type_is is not None and cactual.type_is is not None:
                template_ret_type = template.type_is
                cactual_ret_type = cactual.type_is

            res.extend(infer_constraints(template_ret_type, cactual_ret_type, self.direction))

            if param_spec is None:
                # TODO: Erase template variables if it is generic?
                if (
                    type_state.infer_polymorphic
                    and cactual.variables
                    and not self.skip_neg_op
                    # Technically, the correct inferred type for application of e.g.
                    # Callable[..., T] -> Callable[..., T] (with literal ellipsis), to a generic
                    # like U -> U, should be Callable[..., Any], but if U is a self-type, we can
                    # allow it to leak, to be later bound to self. A bunch of existing code
                    # depends on this old behaviour.
                    and not any(tv.id.is_self() for tv in cactual.variables)
                ):
                    # If the actual callable is generic, infer constraints in the opposite
                    # direction, and indicate to the solver there are extra type variables
                    # to solve for (see more details in mypy/solve.py).
                    res.extend(
                        infer_constraints(
                            cactual, template, neg_op(self.direction), skip_neg_op=True
                        )
                    )
                    extra_tvars = True

                # We can't infer constraints from arguments if the template is Callable[..., T]
                # (with literal '...').
                if not template.is_ellipsis_args:
                    unpack_present = find_unpack_in_list(template.arg_types)
                    # When both ParamSpec and TypeVarTuple are present, things become messy
                    # quickly. For now, we only allow ParamSpec to "capture" TypeVarTuple,
                    # but not vice versa.
                    # TODO: infer more from prefixes when possible.
                    if unpack_present is not None and not cactual.param_spec():
                        # We need to re-normalize args to the form they appear in tuples,
                        # for callables we always pack the suffix inside another tuple.
                        unpack = template.arg_types[unpack_present]
                        assert isinstance(unpack, UnpackType)
                        tuple_type = get_tuple_fallback_from_unpack(unpack)
                        template_types = repack_callable_args(template, tuple_type)
                        actual_types = repack_callable_args(cactual, tuple_type)
                        # Now we can use the same general helper as for tuple types.
                        unpack_constraints = build_constraints_for_simple_unpack(
                            template_types, actual_types, neg_op(self.direction)
                        )
                        res.extend(unpack_constraints)
                    else:
                        # TODO: do we need some special-casing when unpack is present in actual
                        # callable but not in template callable?
                        res.extend(
                            infer_callable_arguments_constraints(template, cactual, self.direction)
                        )
            else:
                prefix = param_spec.prefix
                prefix_len = len(prefix.arg_types)
                cactual_ps = cactual.param_spec()

                if type_state.infer_polymorphic and cactual.variables and not self.skip_neg_op:
                    # Similar logic to the branch above.
                    res.extend(
                        infer_constraints(
                            cactual, template, neg_op(self.direction), skip_neg_op=True
                        )
                    )
                    extra_tvars = True

                # Compare prefixes as well
                cactual_prefix = cactual.copy_modified(
                    arg_types=cactual.arg_types[:prefix_len],
                    arg_kinds=cactual.arg_kinds[:prefix_len],
                    arg_names=cactual.arg_names[:prefix_len],
                )
                res.extend(
                    infer_callable_arguments_constraints(prefix, cactual_prefix, self.direction)
                )

                param_spec_target: Type | None = None
                if not cactual_ps:
                    max_prefix_len = len([k for k in cactual.arg_kinds if k in (ARG_POS, ARG_OPT)])
                    prefix_len = min(prefix_len, max_prefix_len)
                    param_spec_target = Parameters(
                        arg_types=cactual.arg_types[prefix_len:],
                        arg_kinds=cactual.arg_kinds[prefix_len:],
                        arg_names=cactual.arg_names[prefix_len:],
                        variables=cactual.variables if not type_state.infer_polymorphic else [],
                        imprecise_arg_kinds=cactual.imprecise_arg_kinds,
                    )
                else:
                    if len(param_spec.prefix.arg_types) <= len(cactual_ps.prefix.arg_types):
                        param_spec_target = cactual_ps.copy_modified(
                            prefix=Parameters(
                                arg_types=cactual_ps.prefix.arg_types[prefix_len:],
                                arg_kinds=cactual_ps.prefix.arg_kinds[prefix_len:],
                                arg_names=cactual_ps.prefix.arg_names[prefix_len:],
                                imprecise_arg_kinds=cactual_ps.prefix.imprecise_arg_kinds,
                            )
                        )
                if param_spec_target is not None:
                    res.append(Constraint(param_spec, self.direction, param_spec_target))
            if extra_tvars:
                for c in res:
                    c.extra_tvars += cactual.variables
            return res
        elif isinstance(self.actual, AnyType):
            param_spec = template.param_spec()
            any_type = AnyType(TypeOfAny.from_another_any, source_any=self.actual)
            if param_spec is None:
                # FIX what if generic
                res = self.infer_against_any(template.arg_types, self.actual)
            else:
                res = [
                    Constraint(
                        param_spec,
                        SUBTYPE_OF,
                        Parameters([any_type, any_type], [ARG_STAR, ARG_STAR2], [None, None]),
                    )
                ]
            res.extend(infer_constraints(template.ret_type, any_type, self.direction))
            return res
        elif isinstance(self.actual, Overloaded):
            return self.infer_against_overloaded(self.actual, template)
        elif isinstance(self.actual, TypeType):
            return infer_constraints(template.ret_type, self.actual.item, self.direction)
        elif isinstance(self.actual, Instance):
            # Instances with __call__ method defined are considered structural
            # subtypes of Callable with a compatible signature.
            call = mypy.subtypes.find_member(
                "__call__", self.actual, self.actual, is_operator=True
            )
            if call:
                return infer_constraints(template, call, self.direction)
            else:
                return []
        else:
            return []

    def infer_against_overloaded(
        self, overloaded: Overloaded, template: CallableType
    ) -> list[Constraint]:
        # Create constraints by matching an overloaded type against a template.
        # This is tricky to do in general. We cheat by only matching against
        # the first overload item that is callable compatible. This
        # seems to work somewhat well, but we should really use a more
        # reliable technique.
        item = find_matching_overload_item(overloaded, template)
        return infer_constraints(template, item, self.direction)

    def visit_tuple_type(self, template: TupleType) -> list[Constraint]:
        actual = self.actual
        unpack_index = find_unpack_in_list(template.items)
        is_varlength_tuple = (
            isinstance(actual, Instance) and actual.type.fullname == "builtins.tuple"
        )

        if isinstance(actual, TupleType) or is_varlength_tuple:
            res: list[Constraint] = []
            if unpack_index is not None:
                if is_varlength_tuple:
                    # Variadic tuple can be only a supertype of a tuple type, but even if
                    # direction is opposite, inferring something may give better error messages.
                    unpack_type = template.items[unpack_index]
                    assert isinstance(unpack_type, UnpackType)
                    unpacked_type = get_proper_type(unpack_type.type)
                    if isinstance(unpacked_type, TypeVarTupleType):
                        res = [
                            Constraint(type_var=unpacked_type, op=self.direction, target=actual)
                        ]
                    else:
                        assert (
                            isinstance(unpacked_type, Instance)
                            and unpacked_type.type.fullname == "builtins.tuple"
                        )
                        res = infer_constraints(unpacked_type, actual, self.direction)
                    assert isinstance(actual, Instance)  # ensured by is_varlength_tuple == True
                    for i, ti in enumerate(template.items):
                        if i == unpack_index:
                            # This one we just handled above.
                            continue
                        # For Tuple[T, *Ts, S] <: tuple[X, ...] infer also T <: X and S <: X.
                        res.extend(infer_constraints(ti, actual.args[0], self.direction))
                    return res
                else:
                    assert isinstance(actual, TupleType)
                    unpack_constraints = build_constraints_for_simple_unpack(
                        template.items, actual.items, self.direction
                    )
                    actual_items: tuple[Type, ...] = ()
                    template_items: tuple[Type, ...] = ()
                    res.extend(unpack_constraints)
            elif isinstance(actual, TupleType):
                a_unpack_index = find_unpack_in_list(actual.items)
                if a_unpack_index is not None:
                    # The case where template tuple doesn't have an unpack, but actual tuple
                    # has an unpack. We can infer something if actual unpack is a variadic tuple.
                    # Tuple[T, S, U] <: tuple[X, *tuple[Y, ...], Z] => T <: X, S <: Y, U <: Z.
                    a_unpack = actual.items[a_unpack_index]
                    assert isinstance(a_unpack, UnpackType)
                    a_unpacked = get_proper_type(a_unpack.type)
                    if len(actual.items) + 1 <= len(template.items):
                        a_prefix_len = a_unpack_index
                        a_suffix_len = len(actual.items) - a_unpack_index - 1
                        t_prefix, t_middle, t_suffix = split_with_prefix_and_suffix(
                            tuple(template.items), a_prefix_len, a_suffix_len
                        )
                        actual_items = tuple(actual.items[:a_prefix_len])
                        if a_suffix_len:
                            actual_items += tuple(actual.items[-a_suffix_len:])
                        template_items = t_prefix + t_suffix
                        if isinstance(a_unpacked, Instance):
                            assert a_unpacked.type.fullname == "builtins.tuple"
                            for tm in t_middle:
                                res.extend(
                                    infer_constraints(tm, a_unpacked.args[0], self.direction)
                                )
                    else:
                        actual_items = ()
                        template_items = ()
                else:
                    actual_items = tuple(actual.items)
                    template_items = tuple(template.items)
            else:
                return res

            # Cases above will return if actual wasn't a TupleType.
            assert isinstance(actual, TupleType)
            if len(actual_items) == len(template_items):
                if (
                    actual.partial_fallback.type.is_named_tuple
                    and template.partial_fallback.type.is_named_tuple
                ):
                    # For named tuples using just the fallbacks usually gives better results.
                    return res + infer_constraints(
                        template.partial_fallback, actual.partial_fallback, self.direction
                    )
                for i in range(len(template_items)):
                    res.extend(
                        infer_constraints(template_items[i], actual_items[i], self.direction)
                    )
            return res
        elif isinstance(actual, AnyType):
            return self.infer_against_any(template.items, actual)
        else:
            return []

    def visit_typeddict_type(self, template: TypedDictType) -> list[Constraint]:
        actual = self.actual
        if isinstance(actual, TypedDictType):
            res: list[Constraint] = []
            # NOTE: Non-matching keys are ignored. Compatibility is checked
            #       elsewhere so this shouldn't be unsafe.
            for item_name, template_item_type, actual_item_type in template.zip(actual):
                res.extend(infer_constraints(template_item_type, actual_item_type, self.direction))
            return res
        elif isinstance(actual, AnyType):
            return self.infer_against_any(template.items.values(), actual)
        else:
            return []

    def visit_union_type(self, template: UnionType) -> list[Constraint]:
        assert False, (
            "Unexpected UnionType in ConstraintBuilderVisitor"
            " (should have been handled in infer_constraints)"
        )

    def visit_type_alias_type(self, template: TypeAliasType) -> list[Constraint]:
        assert False, f"This should be never called, got {template}"

    def infer_against_any(self, types: Iterable[Type], any_type: AnyType) -> list[Constraint]:
        res: list[Constraint] = []
        # Some items may be things like `*Tuple[*Ts, T]` for example from callable types with
        # suffix after *arg, so flatten them.
        for t in flatten_nested_tuples(types):
            if isinstance(t, UnpackType):
                if isinstance(t.type, TypeVarTupleType):
                    res.append(Constraint(t.type, self.direction, any_type))
                else:
                    unpacked = get_proper_type(t.type)
                    assert isinstance(unpacked, Instance)
                    res.extend(infer_constraints(unpacked, any_type, self.direction))
            else:
                # Note that we ignore variance and simply always use the
                # original direction. This is because for Any targets direction is
                # irrelevant in most cases, see e.g. is_same_constraint().
                res.extend(infer_constraints(t, any_type, self.direction))
        return res

    def visit_overloaded(self, template: Overloaded) -> list[Constraint]:
        if isinstance(self.actual, CallableType):
            items = find_matching_overload_items(template, self.actual)
        else:
            items = template.items
        res: list[Constraint] = []
        for t in items:
            res.extend(infer_constraints(t, self.actual, self.direction))
        return res

    def visit_type_type(self, template: TypeType) -> list[Constraint]:
        if isinstance(self.actual, CallableType):
            return infer_constraints(template.item, self.actual.ret_type, self.direction)
        elif isinstance(self.actual, Overloaded):
            return infer_constraints(template.item, self.actual.items[0].ret_type, self.direction)
        elif isinstance(self.actual, TypeType):
            return infer_constraints(template.item, self.actual.item, self.direction)
        elif isinstance(self.actual, AnyType):
            return infer_constraints(template.item, self.actual, self.direction)
        else:
            return []


def neg_op(op: int) -> int:
    """Map SubtypeOf to SupertypeOf and vice versa."""

    if op == SUBTYPE_OF:
        return SUPERTYPE_OF
    elif op == SUPERTYPE_OF:
        return SUBTYPE_OF
    else:
        raise ValueError(f"Invalid operator {op}")


def find_matching_overload_item(overloaded: Overloaded, template: CallableType) -> CallableType:
    """Disambiguate overload item against a template."""
    items = overloaded.items
    for item in items:
        # Return type may be indeterminate in the template, so ignore it when performing a
        # subtype check.
        if mypy.subtypes.is_callable_compatible(
            item,
            template,
            is_compat=mypy.subtypes.is_subtype,
            is_proper_subtype=False,
            ignore_return=True,
        ):
            return item
    # Fall back to the first item if we can't find a match. This is totally arbitrary --
    # maybe we should just bail out at this point.
    return items[0]


def find_matching_overload_items(
    overloaded: Overloaded, template: CallableType
) -> list[CallableType]:
    """Like find_matching_overload_item, but return all matches, not just the first."""
    items = overloaded.items
    res = []
    for item in items:
        # Return type may be indeterminate in the template, so ignore it when performing a
        # subtype check.
        if mypy.subtypes.is_callable_compatible(
            item,
            template,
            is_compat=mypy.subtypes.is_subtype,
            is_proper_subtype=False,
            ignore_return=True,
        ):
            res.append(item)
    if not res:
        # Falling back to all items if we can't find a match is pretty arbitrary, but
        # it maintains backward compatibility.
        res = items.copy()
    return res


def get_tuple_fallback_from_unpack(unpack: UnpackType) -> TypeInfo:
    """Get builtins.tuple type from available types to construct homogeneous tuples."""
    tp = get_proper_type(unpack.type)
    if isinstance(tp, Instance) and tp.type.fullname == "builtins.tuple":
        return tp.type
    if isinstance(tp, TypeVarTupleType):
        return tp.tuple_fallback.type
    if isinstance(tp, TupleType):
        for base in tp.partial_fallback.type.mro:
            if base.fullname == "builtins.tuple":
                return base
    assert False, "Invalid unpack type"


def repack_callable_args(callable: CallableType, tuple_type: TypeInfo) -> list[Type]:
    """Present callable with star unpack in a normalized form.

    Since positional arguments cannot follow star argument, they are packed in a suffix,
    while prefix is represented as individual positional args. We want to put all in a single
    list with unpack in the middle, and prefix/suffix on the sides (as they would appear
    in e.g. a TupleType).
    """
    if ARG_STAR not in callable.arg_kinds:
        return callable.arg_types
    star_index = callable.arg_kinds.index(ARG_STAR)
    arg_types = callable.arg_types[:star_index]
    star_type = callable.arg_types[star_index]
    suffix_types = []
    if not isinstance(star_type, UnpackType):
        # Re-normalize *args: X -> *args: *tuple[X, ...]
        star_type = UnpackType(Instance(tuple_type, [star_type]))
    else:
        tp = get_proper_type(star_type.type)
        if isinstance(tp, TupleType):
            assert isinstance(tp.items[0], UnpackType)
            star_type = tp.items[0]
            suffix_types = tp.items[1:]
    return arg_types + [star_type] + suffix_types


def build_constraints_for_simple_unpack(
    template_args: list[Type], actual_args: list[Type], direction: int
) -> list[Constraint]:
    """Infer constraints between two lists of types with variadic items.

    This function is only supposed to be called when a variadic item is present in templates.
    If there is no variadic item the actuals, we simply use split_with_prefix_and_suffix()
    and infer prefix <: prefix, suffix <: suffix, variadic <: middle. If there is a variadic
    item in the actuals we need to be more careful, only common prefix/suffix can generate
    constraints, also we can only infer constraints for variadic template item, if template
    prefix/suffix are shorter that actual ones, otherwise there may be partial overlap
    between variadic items, for example if template prefix is longer:

        templates: T1, T2, Ts, Ts, Ts, ...
        actuals:   A1, As, As, As, ...

    Note: this function can only be called for builtin variadic constructors: Tuple and Callable.
    For instances, you should first find correct type argument mapping.
    """
    template_unpack = find_unpack_in_list(template_args)
    assert template_unpack is not None
    template_prefix = template_unpack
    template_suffix = len(template_args) - template_prefix - 1

    t_unpack = None
    res = []

    actual_unpack = find_unpack_in_list(actual_args)
    if actual_unpack is None:
        t_unpack = template_args[template_unpack]
        if template_prefix + template_suffix > len(actual_args):
            # These can't be subtypes of each-other, return fast.
            assert isinstance(t_unpack, UnpackType)
            if isinstance(t_unpack.type, TypeVarTupleType):
                # Set TypeVarTuple to empty to improve error messages.
                return [
                    Constraint(
                        t_unpack.type, direction, TupleType([], t_unpack.type.tuple_fallback)
                    )
                ]
            else:
                return []
        common_prefix = template_prefix
        common_suffix = template_suffix
    else:
        actual_prefix = actual_unpack
        actual_suffix = len(actual_args) - actual_prefix - 1
        common_prefix = min(template_prefix, actual_prefix)
        common_suffix = min(template_suffix, actual_suffix)
        if actual_prefix >= template_prefix and actual_suffix >= template_suffix:
            # This is the only case where we can guarantee there will be no partial overlap
            # (note however partial overlap is OK for variadic tuples, it is handled below).
            t_unpack = template_args[template_unpack]

    # Handle constraints from prefixes/suffixes first.
    start, middle, end = split_with_prefix_and_suffix(
        tuple(actual_args), common_prefix, common_suffix
    )
    for t, a in zip(template_args[:common_prefix], start):
        res.extend(infer_constraints(t, a, direction))
    if common_suffix:
        for t, a in zip(template_args[-common_suffix:], end):
            res.extend(infer_constraints(t, a, direction))

    if t_unpack is not None:
        # Add constraint(s) for variadic item when possible.
        assert isinstance(t_unpack, UnpackType)
        tp = get_proper_type(t_unpack.type)
        if isinstance(tp, Instance) and tp.type.fullname == "builtins.tuple":
            # Homogeneous case *tuple[T, ...] <: [X, Y, Z, ...].
            for a in middle:
                # TODO: should we use union instead of join here?
                if not isinstance(a, UnpackType):
                    res.extend(infer_constraints(tp.args[0], a, direction))
                else:
                    a_tp = get_proper_type(a.type)
                    # This is the case *tuple[T, ...] <: *tuple[A, ...].
                    if isinstance(a_tp, Instance) and a_tp.type.fullname == "builtins.tuple":
                        res.extend(infer_constraints(tp.args[0], a_tp.args[0], direction))
        elif isinstance(tp, TypeVarTupleType):
            res.append(Constraint(tp, direction, TupleType(list(middle), tp.tuple_fallback)))
    elif actual_unpack is not None:
        # A special case for a variadic tuple unpack, we simply infer T <: X from
        # Tuple[..., *tuple[T, ...], ...] <: Tuple[..., *tuple[X, ...], ...].
        actual_unpack_type = actual_args[actual_unpack]
        assert isinstance(actual_unpack_type, UnpackType)
        a_unpacked = get_proper_type(actual_unpack_type.type)
        if isinstance(a_unpacked, Instance) and a_unpacked.type.fullname == "builtins.tuple":
            t_unpack = template_args[template_unpack]
            assert isinstance(t_unpack, UnpackType)
            tp = get_proper_type(t_unpack.type)
            if isinstance(tp, Instance) and tp.type.fullname == "builtins.tuple":
                res.extend(infer_constraints(tp.args[0], a_unpacked.args[0], direction))
    return res


def infer_directed_arg_constraints(left: Type, right: Type, direction: int) -> list[Constraint]:
    """Infer constraints between two arguments using direction between original callables."""
    if isinstance(left, (ParamSpecType, UnpackType)) or isinstance(
        right, (ParamSpecType, UnpackType)
    ):
        # This avoids bogus constraints like T <: P.args
        # TODO: can we infer something useful for *T vs P?
        return []
    if direction == SUBTYPE_OF:
        # We invert direction to account for argument contravariance.
        return infer_constraints(left, right, neg_op(direction))
    else:
        return infer_constraints(right, left, neg_op(direction))


def infer_callable_arguments_constraints(
    template: NormalizedCallableType | Parameters,
    actual: NormalizedCallableType | Parameters,
    direction: int,
) -> list[Constraint]:
    """Infer constraints between argument types of two callables.

    This function essentially extracts four steps from are_parameters_compatible() in
    subtypes.py that involve subtype checks between argument types. We keep the argument
    matching logic, but ignore various strictness flags present there, and checks that
    do not involve subtyping. Then in place of every subtype check we put an infer_constraints()
    call for the same types.
    """
    res = []
    if direction == SUBTYPE_OF:
        left, right = template, actual
    else:
        left, right = actual, template
    left_star = left.var_arg()
    left_star2 = left.kw_arg()
    right_star = right.var_arg()
    right_star2 = right.kw_arg()

    # Numbering of steps below matches the one in are_parameters_compatible() for convenience.
    # Phase 1a: compare star vs star arguments.
    if left_star is not None and right_star is not None:
        res.extend(infer_directed_arg_constraints(left_star.typ, right_star.typ, direction))
    if left_star2 is not None and right_star2 is not None:
        res.extend(infer_directed_arg_constraints(left_star2.typ, right_star2.typ, direction))

    # Phase 1b: compare left args with corresponding non-star right arguments.
    for right_arg in right.formal_arguments():
        left_arg = mypy.typeops.callable_corresponding_argument(left, right_arg)
        if left_arg is None:
            continue
        res.extend(infer_directed_arg_constraints(left_arg.typ, right_arg.typ, direction))

    # Phase 1c: compare left args with right *args.
    if right_star is not None:
        right_by_position = right.try_synthesizing_arg_from_vararg(None)
        assert right_by_position is not None
        i = right_star.pos
        assert i is not None
        while i < len(left.arg_kinds) and left.arg_kinds[i].is_positional():
            left_by_position = left.argument_by_position(i)
            assert left_by_position is not None
            res.extend(
                infer_directed_arg_constraints(
                    left_by_position.typ, right_by_position.typ, direction
                )
            )
            i += 1

    # Phase 1d: compare left args with right **kwargs.
    if right_star2 is not None:
        right_names = {name for name in right.arg_names if name is not None}
        left_only_names = set()
        for name, kind in zip(left.arg_names, left.arg_kinds):
            if name is None or kind.is_star() or name in right_names:
                continue
            left_only_names.add(name)

        right_by_name = right.try_synthesizing_arg_from_kwarg(None)
        assert right_by_name is not None
        for name in left_only_names:
            left_by_name = left.argument_by_name(name)
            assert left_by_name is not None
            res.extend(
                infer_directed_arg_constraints(left_by_name.typ, right_by_name.typ, direction)
            )
    return res


def filter_imprecise_kinds(cs: list[Constraint]) -> list[Constraint]:
    """For each ParamSpec remove all imprecise constraints, if at least one precise available."""
    have_precise = set()
    for c in cs:
        if not isinstance(c.origin_type_var, ParamSpecType):
            continue
        if (
            isinstance(c.target, ParamSpecType)
            or isinstance(c.target, Parameters)
            and not c.target.imprecise_arg_kinds
        ):
            have_precise.add(c.type_var)
    new_cs = []
    for c in cs:
        if not isinstance(c.origin_type_var, ParamSpecType) or c.type_var not in have_precise:
            new_cs.append(c)
        if not isinstance(c.target, Parameters) or not c.target.imprecise_arg_kinds:
            new_cs.append(c)
    return new_cs



"""
Cloud Settings - Temel Sayfa Sınıfı
Tüm ayar sayfaları için temel sınıf
"""

import logging
from typing import Dict, Any
from pathlib import Path
from PyQt6.QtWidgets import *
from PyQt6.QtCore import *
from PyQt6.QtGui import *

class BasePage(QWidget):
    """Ayar sayfası temel sınıfı"""
    
    settings_changed = pyqtSignal(str, dict)  # category, settings
    
    def __init__(self, category: str, title: str, icon: str = "⚙️", kernel=None, preview_manager=None):
        super().__init__()
        self.category = category
        self.title = title
        self.icon = icon
        self.kernel = kernel
        self.preview_manager = preview_manager
        self.logger = logging.getLogger(f"SettingsPage.{category}")
        
        self.settings = {}
        self.widgets = {}
        
        self.setup_ui()
        self.setup_connections()
    
    def setup_ui(self):
        """UI kurulumu - alt sınıflarda override edilecek"""
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(20)
        
        # Scroll area
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setFrameShape(QFrame.Shape.NoFrame)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        
        # İçerik widget'ı
        self.content_widget = QWidget()
        self.content_layout = QVBoxLayout(self.content_widget)
        self.content_layout.setContentsMargins(0, 0, 20, 0)  # Sağ tarafta scroll bar için boşluk
        self.content_layout.setSpacing(20)
        
        scroll.setWidget(self.content_widget)
        layout.addWidget(scroll)
        
        # Scroll area stili
        scroll.setStyleSheet("""
            QScrollArea {
                background-color: transparent;
                border: none;
            }
            
            QScrollBar:vertical {
                background-color: #f8f9fa;
                width: 12px;
                border-radius: 6px;
                margin: 0;
            }
            
            QScrollBar::handle:vertical {
                background-color: #dee2e6;
                border-radius: 6px;
                min-height: 20px;
            }
            
            QScrollBar::handle:vertical:hover {
                background-color: #adb5bd;
            }
            
            QScrollBar::add-line:vertical,
            QScrollBar::sub-line:vertical {
                height: 0;
            }
        """)
    
    def setup_connections(self):
        """Sinyal bağlantıları - alt sınıflarda override edilebilir"""
        pass
    
    def add_group(self, title: str) -> QGroupBox:
        """Grup kutusu ekle"""
        from ..widgets import ModernGroupBox
        
        group = ModernGroupBox(title)
        self.content_layout.addWidget(group)
        return group
    
    def add_setting_row(self, parent_layout, label_text: str, widget: QWidget, description: str = ""):
        """Ayar satırı ekle"""
        row_widget = QWidget()
        row_layout = QHBoxLayout(row_widget)
        row_layout.setContentsMargins(0, 0, 0, 0)
        
        # Sol taraf - etiket ve açıklama
        left_widget = QWidget()
        left_layout = QVBoxLayout(left_widget)
        left_layout.setContentsMargins(0, 0, 0, 0)
        left_layout.setSpacing(2)
        
        label = QLabel(label_text)
        label.setStyleSheet("""
            QLabel {
                font-size: 14px;
                font-weight: 600;
                color: #212529;
            }
        """)
        left_layout.addWidget(label)
        
        if description:
            desc_label = QLabel(description)
            desc_label.setStyleSheet("""
                QLabel {
                    font-size: 12px;
                    color: #6c757d;
                }
            """)
            desc_label.setWordWrap(True)
            left_layout.addWidget(desc_label)
        
        left_layout.addStretch()
        row_layout.addWidget(left_widget, 1)
        
        # Sağ taraf - widget
        row_layout.addWidget(widget)
        
        parent_layout.addWidget(row_widget)
        return row_widget
    
    def add_checkbox(self, parent_layout, text: str, key: str, description: str = ""):
        """Checkbox ekle"""
        from ..widgets import ModernCheckBox
        
        checkbox = ModernCheckBox(text)
        checkbox.toggled.connect(lambda checked: self.on_setting_changed(key, checked))
        
        self.widgets[key] = checkbox
        
        if description:
            self.add_setting_row(parent_layout, "", checkbox, description)
        else:
            parent_layout.addWidget(checkbox)
        
        return checkbox
    
    def add_combobox(self, parent_layout, label: str, key: str, items: list, description: str = ""):
        """ComboBox ekle"""
        from ..widgets import ModernComboBox
        
        combo = ModernComboBox()
        combo.addItems(items)
        combo.currentTextChanged.connect(lambda text: self.on_setting_changed(key, text.lower()))
        
        self.widgets[key] = combo
        self.add_setting_row(parent_layout, label, combo, description)
        
        return combo
    
    def add_slider(self, parent_layout, label: str, key: str, min_val: int, max_val: int, default: int, description: str = ""):
        """Slider ekle"""
        from ..widgets import ModernSlider
        
        slider = ModernSlider(min_val, max_val, default)
        slider.value_changed.connect(lambda value: self.on_setting_changed(key, value))
        
        self.widgets[key] = slider
        self.add_setting_row(parent_layout, label, slider, description)
        
        return slider
    
    def add_spinbox(self, parent_layout, label: str, key: str, min_val: int, max_val: int, default: int, suffix: str = "", description: str = ""):
        """SpinBox ekle"""
        from ..widgets import ModernSpinBox
        
        spinbox = ModernSpinBox()
        spinbox.setRange(min_val, max_val)
        spinbox.setValue(default)
        if suffix:
            spinbox.setSuffix(suffix)
        spinbox.valueChanged.connect(lambda value: self.on_setting_changed(key, value))
        
        self.widgets[key] = spinbox
        self.add_setting_row(parent_layout, label, spinbox, description)
        
        return spinbox
    
    def add_color_picker(self, parent_layout, label: str, key: str, default_color: str = "#007bff", description: str = ""):
        """Renk seçici ekle"""
        from ..widgets import ColorPickerButton
        
        color_picker = ColorPickerButton(default_color)
        color_picker.color_changed.connect(lambda color: self.on_setting_changed(key, color))
        
        self.widgets[key] = color_picker
        self.add_setting_row(parent_layout, label, color_picker, description)
        
        return color_picker
    
    def add_file_picker(self, parent_layout, label: str, key: str, file_filter: str = "", description: str = ""):
        """Dosya seçici ekle"""
        file_widget = QWidget()
        file_layout = QHBoxLayout(file_widget)
        file_layout.setContentsMargins(0, 0, 0, 0)
        
        line_edit = QLineEdit()
        line_edit.setPlaceholderText("Dosya seçin...")
        line_edit.textChanged.connect(lambda text: self.on_setting_changed(key, text))
        
        browse_btn = QPushButton("Gözat")
        browse_btn.clicked.connect(lambda: self._browse_file(line_edit, file_filter))
        
        file_layout.addWidget(line_edit, 1)
        file_layout.addWidget(browse_btn)
        
        self.widgets[key] = line_edit
        self.add_setting_row(parent_layout, label, file_widget, description)
        
        return line_edit
    
    def _browse_file(self, line_edit: QLineEdit, file_filter: str):
        """Dosya gözat"""
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Dosya Seç",
            line_edit.text() or str(Path.home()),
            file_filter
        )
        
        if file_path:
            line_edit.setText(file_path)
    
    def on_setting_changed(self, key: str, value):
        """Ayar değişti"""
        self.settings[key] = value
        
        # Canlı önizleme güncelle
        if self.preview_manager:
            self.preview_manager.update_preview(self.category, self.settings)
        
        # Sinyal gönder
        self.settings_changed.emit(self.category, self.settings.copy())
        
        self.logger.debug(f"Setting changed: {key} = {value}")
    
    def load_settings(self, settings: Dict[str, Any]):
        """Ayarları yükle"""
        self.settings = settings.copy()
        self.update_ui()
        
        self.logger.debug(f"Settings loaded: {len(settings)} items")
    
    def update_ui(self):
        """UI'yi ayarlara göre güncelle - alt sınıflarda override edilecek"""
        for key, widget in self.widgets.items():
            if key in self.settings:
                value = self.settings[key]
                
                try:
                    if isinstance(widget, QCheckBox):
                        widget.setChecked(bool(value))
                    elif isinstance(widget, QComboBox):
                        # Değeri bul ve seç
                        for i in range(widget.count()):
                            if widget.itemText(i).lower() == str(value).lower():
                                widget.setCurrentIndex(i)
                                break
                    elif isinstance(widget, QSlider):
                        widget.setValue(int(value))
                    elif isinstance(widget, QSpinBox):
                        widget.setValue(int(value))
                    elif isinstance(widget, QLineEdit):
                        widget.setText(str(value))
                    elif hasattr(widget, 'setValue'):
                        widget.setValue(value)
                    elif hasattr(widget, 'set_color'):
                        widget.set_color(str(value))
                        
                except Exception as e:
                    self.logger.warning(f"Failed to update widget {key}: {e}")
    
    def get_settings(self) -> Dict[str, Any]:
        """Mevcut ayarları al"""
        return self.settings.copy()
    
    def add_spacer(self):
        """Boşluk ekle"""
        self.content_layout.addStretch()
    
    def add_separator(self):
        """Ayırıcı çizgi ekle"""
        line = QFrame()
        line.setFrameShape(QFrame.Shape.HLine)
        line.setFrameShadow(QFrame.Shadow.Sunken)
        line.setStyleSheet("""
            QFrame {
                color: #e9ecef;
                background-color: #e9ecef;
                border: none;
                height: 1px;
            }
        """)
        self.content_layout.addWidget(line) 


"""
Cloud FilePicker - PyCloud OS Dosya Seçim Penceresi
VFS ile entegre çalışan güvenli dosya seçici
"""

import os
import sys
import logging
from pathlib import Path
from typing import List, Optional, Dict, Callable
from datetime import datetime

# PyQt6 import
try:
    from PyQt6.QtWidgets import *
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
    PYQT_AVAILABLE = True
except ImportError:
    PYQT_AVAILABLE = False

class FilePickerMode:
    """Dosya seçici modları"""
    OPEN_FILE = "open_file"
    SAVE_FILE = "save_file"
    SELECT_DIRECTORY = "select_directory"
    MULTIPLE_FILES = "multiple_files"

class FilePickerFilter:
    """Dosya filtreleri"""
    ALL_FILES = ("Tüm Dosyalar", "*")
    TEXT_FILES = ("Metin Dosyaları", "*.txt;*.md;*.py;*.json;*.log")
    PYTHON_FILES = ("Python Dosyaları", "*.py;*.pyw;*.pyi")
    IMAGES = ("Resim Dosyaları", "*.jpg;*.jpeg;*.png;*.gif;*.bmp")
    DOCUMENTS = ("Belgeler", "*.pdf;*.doc;*.docx;*.odt")
    ARCHIVES = ("Arşiv Dosyaları", "*.zip;*.tar;*.gz;*.rar")
    CODE_FILES = ("Kod Dosyaları", "*.py;*.js;*.html;*.css;*.json;*.xml;*.yaml;*.yml")
    
    @staticmethod
    def custom(name: str, extensions: str):
        """Özel filtre oluştur"""
        return (name, extensions)

if PYQT_AVAILABLE:
    class CloudFilePicker(QDialog):
        """PyCloud OS Dosya Seçici"""
        
        # Sinyaller
        fileSelected = pyqtSignal(str)  # Dosya seçildi
        filesSelected = pyqtSignal(list)  # Çoklu dosya seçildi
        directorySelected = pyqtSignal(str)  # Dizin seçildi
        
        def __init__(self, parent=None, kernel=None):
            super().__init__(parent)
            self.kernel = kernel
            self.logger = logging.getLogger("CloudFilePicker")
            
            # VFS referansı
            self.vfs = None
            if self.kernel:
                try:
                    self.vfs = self.kernel.get_module("vfs")
                except:
                    self.logger.warning("VFS module not available")
            
            # UI bileşenleri
            self.path_bar = None
            self.file_list = None
            self.file_name_edit = None
            self.filter_combo = None
            self.button_box = None
            
            # Durum
            self.current_path = "/home"
            self.mode = FilePickerMode.OPEN_FILE
            self.filters = [FilePickerFilter.ALL_FILES]
            self.selected_files = []
            self.app_id = "unknown"
            self.multi_select = False
            
            # Tema
            self.is_dark_mode = self._detect_dark_mode()
            
            self._init_ui()
            self._connect_signals()
            self._apply_theme()
            
        def _detect_dark_mode(self) -> bool:
            """Sistem temasını tespit et"""
            try:
                if self.kernel:
                    config = self.kernel.get_module("config")
                    if config:
                        theme_config = config.get("theme", {})
                        return theme_config.get("dark_mode", False)
            except:
                pass
            return False
        
        def _init_ui(self):
            """UI bileşenlerini oluştur"""
            self.setWindowTitle("Dosya Seç")
            self.setFixedSize(800, 600)
            
            # Ana layout
            layout = QVBoxLayout(self)
            layout.setContentsMargins(10, 10, 10, 10)
            layout.setSpacing(8)
            
            # Üst panel - navigation
            nav_layout = QHBoxLayout()
            
            # Geri/İleri butonları
            self.back_btn = QPushButton("←")
            self.back_btn.setFixedSize(32, 32)
            self.back_btn.setToolTip("Geri")
            nav_layout.addWidget(self.back_btn)
            
            self.forward_btn = QPushButton("→")
            self.forward_btn.setFixedSize(32, 32)
            self.forward_btn.setToolTip("İleri")
            nav_layout.addWidget(self.forward_btn)
            
            # Ana dizin butonu
            self.home_btn = QPushButton("🏠")
            self.home_btn.setFixedSize(32, 32)
            self.home_btn.setToolTip("Ana Dizin")
            nav_layout.addWidget(self.home_btn)
            
            # Yol çubuğu
            self.path_bar = QLineEdit()
            self.path_bar.setPlaceholderText("Dosya yolu...")
            nav_layout.addWidget(self.path_bar)
            
            layout.addLayout(nav_layout)
            
            # Orta panel - dosya listesi
            self.file_list = QListWidget()
            self.file_list.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
            self.file_list.setIconSize(QSize(24, 24))
            layout.addWidget(self.file_list)
            
            # Alt panel - dosya adı ve filtreler
            bottom_layout = QGridLayout()
            
            # Dosya adı
            bottom_layout.addWidget(QLabel("Dosya adı:"), 0, 0)
            self.file_name_edit = QLineEdit()
            bottom_layout.addWidget(self.file_name_edit, 0, 1, 1, 2)
            
            # Filtre
            bottom_layout.addWidget(QLabel("Dosya türü:"), 1, 0)
            self.filter_combo = QComboBox()
            bottom_layout.addWidget(self.filter_combo, 1, 1, 1, 2)
            
            layout.addLayout(bottom_layout)
            
            # Buton panel
            self.button_box = QDialogButtonBox()
            self.ok_btn = self.button_box.addButton("Tamam", QDialogButtonBox.ButtonRole.AcceptRole)
            self.cancel_btn = self.button_box.addButton("İptal", QDialogButtonBox.ButtonRole.RejectRole)
            layout.addWidget(self.button_box)
            
        def _connect_signals(self):
            """Sinyalleri bağla"""
            # Navigation
            self.back_btn.clicked.connect(self._go_back)
            self.forward_btn.clicked.connect(self._go_forward)
            self.home_btn.clicked.connect(self._go_home)
            self.path_bar.returnPressed.connect(self._navigate_to_path)
            
            # File list
            self.file_list.itemDoubleClicked.connect(self._on_item_double_clicked)
            self.file_list.itemSelectionChanged.connect(self.selection_changed)
            
            # Filter
            self.filter_combo.currentTextChanged.connect(self._apply_filter)
            
            # Buttons
            self.button_box.accepted.connect(self.accept_selection)
            self.button_box.rejected.connect(self.reject)
            
        def _apply_theme(self):
            """Tema uygula"""
            if self.is_dark_mode:
                # Koyu tema
                self.setStyleSheet("""
                    QDialog {
                        background-color: #2b2b2b;
                        color: white;
                    }
                    QListWidget {
                        background-color: #353535;
                        border: 1px solid #555;
                        border-radius: 4px;
                    }
                    QListWidget::item {
                        padding: 8px;
                        border-bottom: 1px solid #444;
                    }
                    QListWidget::item:selected {
                        background-color: #0078d4;
                    }
                    QLineEdit, QComboBox {
                        background-color: #353535;
                        border: 1px solid #555;
                        border-radius: 4px;
                        padding: 6px;
                    }
                    QPushButton {
                        background-color: #404040;
                        border: 1px solid #555;
                        border-radius: 4px;
                        padding: 6px 12px;
                    }
                    QPushButton:hover {
                        background-color: #4a4a4a;
                    }
                    QPushButton:pressed {
                        background-color: #333;
                    }
                """)
            else:
                # Açık tema
                self.setStyleSheet("""
                    QDialog {
                        background-color: white;
                        color: black;
                    }
                    QListWidget {
                        background-color: white;
                        border: 1px solid #ccc;
                        border-radius: 4px;
                    }
                    QListWidget::item {
                        padding: 8px;
                        border-bottom: 1px solid #eee;
                    }
                    QListWidget::item:selected {
                        background-color: #0078d4;
                        color: white;
                    }
                    QLineEdit, QComboBox {
                        background-color: white;
                        border: 1px solid #ccc;
                        border-radius: 4px;
                        padding: 6px;
                    }
                    QPushButton {
                        background-color: #f0f0f0;
                        border: 1px solid #ccc;
                        border-radius: 4px;
                        padding: 6px 12px;
                    }
                    QPushButton:hover {
                        background-color: #e0e0e0;
                    }
                """)
        
        def setup(self, mode: str, filters: List[tuple] = None, app_id: str = "unknown", 
                 multi_select: bool = False, initial_path: str = None):
            """FilePicker'ı yapılandır"""
            self.mode = mode
            self.app_id = app_id
            self.multi_select = multi_select
            
            # Çoklu seçim
            if multi_select:
                self.file_list.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
            
            # Filtreler
            if filters:
                self.filters = filters
            
            self.filter_combo.clear()
            for name, ext in self.filters:
                self.filter_combo.addItem(f"{name} ({ext})")
            
            # Başlangıç dizini
            if initial_path and self._is_path_allowed(initial_path):
                self.current_path = initial_path
            else:
                self.current_path = self._get_default_path()
            
            # UI güncellemeleri
            self._update_title()
            self._navigate_to(self.current_path)
            
        def _update_title(self):
            """Pencere başlığını güncelle"""
            titles = {
                FilePickerMode.OPEN_FILE: "Dosya Aç",
                FilePickerMode.SAVE_FILE: "Dosyayı Kaydet",
                FilePickerMode.SELECT_DIRECTORY: "Dizin Seç",
                FilePickerMode.MULTIPLE_FILES: "Dosyalar Seç"
            }
            self.setWindowTitle(titles.get(self.mode, "Dosya Seç"))
            
        def _get_default_path(self) -> str:
            """Varsayılan başlangıç yolu"""
            if self.vfs:
                allowed_paths = self.vfs.list_allowed_paths(self.app_id)
                if "/home" in allowed_paths:
                    return "/home"
                elif allowed_paths:
                    return allowed_paths[0]
            return "/home"
        
        def _is_path_allowed(self, path: str) -> bool:
            """Yola erişim izni var mı?"""
            if not self.vfs:
                return True
            return self.vfs.validate_path(path, self.app_id)
        
        def _navigate_to(self, path: str):
            """Belirtilen dizine git"""
            try:
                if not self._is_path_allowed(path):
                    self.logger.warning(f"Access denied to path: {path}")
                    return
                
                self.current_path = path
                self.path_bar.setText(path)
                self._load_directory()
                
            except Exception as e:
                self.logger.error(f"Navigation error: {e}")
        
        def _load_directory(self):
            """Dizin içeriğini yükle"""
            try:
                self.file_list.clear()
                
                # VFS üzerinden dizin listesi al
                if self.vfs:
                    real_path = self.vfs.resolve_path(self.current_path)
                    if not real_path or not os.path.exists(real_path):
                        return
                    
                    real_path_obj = Path(real_path)
                else:
                    # Fallback - direkt sistem
                    real_path_obj = Path(self.current_path)
                
                # Üst dizin (".." seçeneği)
                if self.current_path != "/":
                    parent_item = QListWidgetItem("📁 ..")
                    parent_item.setData(Qt.ItemDataRole.UserRole, "directory")
                    self.file_list.addItem(parent_item)
                
                # Dizin ve dosyaları listele
                if real_path_obj.exists():
                    items = []
                    for item in real_path_obj.iterdir():
                        try:
                            if item.is_dir():
                                items.append((item.name, "directory"))
                            else:
                                items.append((item.name, "file"))
                        except:
                            continue
                    
                    # Sırala: dizinler önce, sonra dosyalar
                    items.sort(key=lambda x: (x[1] != "directory", x[0].lower()))
                    
                    for name, item_type in items:
                        icon = "📁" if item_type == "directory" else self._get_file_icon(name)
                        list_item = QListWidgetItem(f"{icon} {name}")
                        list_item.setData(Qt.ItemDataRole.UserRole, item_type)
                        self.file_list.addItem(list_item)
                
            except Exception as e:
                self.logger.error(f"Failed to load directory: {e}")
        
        def _get_file_icon(self, filename: str) -> str:
            """Dosya türüne göre ikon"""
            ext = Path(filename).suffix.lower()
            icons = {
                '.txt': '📄', '.md': '📝', '.py': '🐍', '.js': '📜',
                '.html': '🌐', '.css': '🎨', '.json': '📋',
                '.jpg': '🖼️', '.jpeg': '🖼️', '.png': '🖼️', '.gif': '🖼️',
                '.mp3': '🎵', '.wav': '🎵', '.mp4': '🎬', '.avi': '🎬',
                '.pdf': '📕', '.zip': '📦', '.tar': '📦', '.gz': '📦'
            }
            return icons.get(ext, '📄')
        
        def _navigate_to_path(self):
            """Yol çubuğundan navigasyon"""
            path = self.path_bar.text().strip()
            if path:
                self._navigate_to(path)
        
        def _go_back(self):
            """Üst dizine git"""
            if self.current_path != "/":
                parent = str(Path(self.current_path).parent)
                if parent != self.current_path:
                    self._navigate_to(parent)
        
        def _go_forward(self):
            """İleri git (şimdilik placeholder)"""
            pass
        
        def _go_home(self):
            """Ana dizine git"""
            home_path = self._get_default_path()
            self._navigate_to(home_path)
        
        def _on_item_double_clicked(self, item: QListWidgetItem):
            """Öğeye çift tıklama"""
            item_type = item.data(Qt.ItemDataRole.UserRole)
            item_text = item.text()
            
            # İkon prefixi kaldır
            name = item_text.split(" ", 1)[1] if " " in item_text else item_text
            
            if item_type == "directory":
                if name == "..":
                    self._go_back()
                else:
                    new_path = str(Path(self.current_path) / name)
                    self._navigate_to(new_path)
            elif self.mode in [FilePickerMode.OPEN_FILE, FilePickerMode.MULTIPLE_FILES]:
                # Dosya seçildi, dialog'u kapat
                self.accept()
        
        def selection_changed(self):
            """Seçim değiştiğinde çağrılır"""
            try:
                if self.mode == FilePickerMode.MULTIPLE_FILES:
                    # Multiple selection handling
                    selected_items = self.file_list.selectedItems()
                    selected_count = len(selected_items)
                    
                    self.selection_info.setText(f"{selected_count} dosya seçildi")
                    self.action_button.setEnabled(selected_count > 0)
                    
                    # Seçili dosyaları sakla
                    self.selected_files = []
                    for item in selected_items:
                        filename = item.text()
                        if not filename.startswith("📁"):  # Klasör değilse
                            file_path = str(Path(self.current_path) / filename)
                            self.selected_files.append(file_path)
                    
                else:
                    # Single selection handling
                    selected_items = self.file_list.selectedItems()
                    if selected_items:
                        filename = selected_items[0].text()
                        if filename.startswith("📁"):  # Klasör
                            self.action_button.setEnabled(self.mode == FilePickerMode.SELECT_DIRECTORY)
                        else:  # Dosya
                            self.action_button.setEnabled(self.mode in [FilePickerMode.OPEN_FILE])
                            if hasattr(self, 'filename_input'):
                                self.filename_input.setText(filename)
                    else:
                        self.action_button.setEnabled(False)
                    
            except Exception as e:
                self.logger.error(f"Selection changed error: {e}")
        
        def accept_selection(self):
            """Seçimi kabul et"""
            try:
                if self.mode == FilePickerMode.MULTIPLE_FILES:
                    # Multiple files return
                    if hasattr(self, 'selected_files') and self.selected_files:
                        self.selected_paths = self.selected_files
                        self.accept()
                    
                elif self.mode == FilePickerMode.SAVE_FILE:
                    # Save file handling
                    if hasattr(self, 'filename_input'):
                        filename = self.filename_input.text().strip()
                        if filename:
                            self.selected_path = str(Path(self.current_path) / filename)
                            self.accept()
                            
                elif self.mode == FilePickerMode.OPEN_FILE:
                    # Open file handling
                    selected_items = self.file_list.selectedItems()
                    if selected_items:
                        filename = selected_items[0].text()
                        if not filename.startswith("📁"):
                            self.selected_path = str(Path(self.current_path) / filename)
                            self.accept()
                            
                elif self.mode == FilePickerMode.SELECT_DIRECTORY:
                    # Directory selection
                    self.selected_path = self.current_path
                    self.accept()
                    
            except Exception as e:
                self.logger.error(f"Accept selection error: {e}")
        
        def get_selected_files(self) -> List[str]:
            """Seçili dosyaları al (multiple mode için)"""
            if hasattr(self, 'selected_paths'):
                return self.selected_paths
            elif hasattr(self, 'selected_path'):
                return [self.selected_path]
            else:
                return []
        
        def _apply_filter(self):
            """Filtre uygula"""
            # Şimdilik placeholder - gelecekte dosya filtreleme eklenebilir
            pass

def show_file_picker(mode: str = FilePickerMode.OPEN_FILE, 
                    filters: List[tuple] = None,
                    app_id: str = "unknown",
                    multi_select: bool = False,
                    initial_path: str = None,
                    parent=None,
                    kernel=None) -> Optional[str]:
    """Dosya seçici göster"""
    if not PYQT_AVAILABLE:
        return None
    
    try:
        picker = CloudFilePicker(parent=parent, kernel=kernel)
        picker.setup(mode=mode, filters=filters, app_id=app_id, 
                    multi_select=multi_select, initial_path=initial_path)
        
        # Modal dialog olarak göster
        if picker.exec() == QDialog.DialogCode.Accepted:
            if mode == FilePickerMode.SELECT_DIRECTORY:
                return picker.current_path
            else:
                filename = picker.file_name_edit.text().strip()
                if filename:
                    return str(Path(picker.current_path) / filename)
        
        return None
        
    except Exception as e:
        logging.getLogger("CloudFilePicker").error(f"FilePicker error: {e}")
        return None

# Convenience functions
def open_file_dialog(app_id: str = "unknown", filters: List[tuple] = None, 
                    parent=None, kernel=None) -> Optional[str]:
    """Dosya açma dialog'u"""
    return show_file_picker(
        mode=FilePickerMode.OPEN_FILE,
        filters=filters or [FilePickerFilter.ALL_FILES],
        app_id=app_id,
        parent=parent,
        kernel=kernel
    )

def save_file_dialog(app_id: str = "unknown", filters: List[tuple] = None,
                    parent=None, kernel=None) -> Optional[str]:
    """Dosya kaydetme dialog'u"""
    return show_file_picker(
        mode=FilePickerMode.SAVE_FILE,
        filters=filters or [FilePickerFilter.ALL_FILES],
        app_id=app_id,
        parent=parent,
        kernel=kernel
    )

def select_directory_dialog(app_id: str = "unknown", parent=None, kernel=None) -> Optional[str]:
    """Dizin seçme dialog'u"""
    return show_file_picker(
        mode=FilePickerMode.SELECT_DIRECTORY,
        app_id=app_id,
        parent=parent,
        kernel=kernel
    )

def select_multiple_files_dialog(app_id: str = "unknown", filters: List[tuple] = None,
                                parent=None, kernel=None) -> Optional[List[str]]:
    """Çoklu dosya seçme dialog'u"""
    if not PYQT_AVAILABLE:
        return None
    
    try:
        from PyQt6.QtWidgets import QApplication
        
        # App instance kontrolü
        app = QApplication.instance()
        if app is None:
            app = QApplication([])
        
        # Custom FilePicker window oluştur
        picker = CloudFilePicker(
            parent=parent,
            kernel=kernel
        )
        picker.setup(
            mode=FilePickerMode.MULTIPLE_FILES,
            filters=filters or [FilePickerFilter.ALL_FILES],
            app_id=app_id,
            multi_select=True
        )
        
        if picker.exec() == QDialog.DialogCode.Accepted:
            return picker.get_selected_files()
        
        return None
        
    except Exception as e:
        logging.getLogger("CloudFilePicker").error(f"Multiple files picker error: {e}")
        return None

def setup_ui(self):
    """UI kurulumu"""
    layout = QVBoxLayout()
    layout.setContentsMargins(12, 12, 12, 12)
    layout.setSpacing(8)
    
    # Başlık
    title_text = {
        FilePickerMode.OPEN_FILE: "📁 Dosya Aç",
        FilePickerMode.SAVE_FILE: "💾 Dosya Kaydet", 
        FilePickerMode.SELECT_DIRECTORY: "📂 Klasör Seç",
        FilePickerMode.MULTIPLE_FILES: "📋 Çoklu Dosya Seç"
    }
    
    title = QLabel(title_text.get(self.mode, "📁 Dosya İşlemi"))
    title.setAlignment(Qt.AlignmentFlag.AlignCenter)
    title.setStyleSheet("font-size: 16px; font-weight: bold; margin-bottom: 10px;")
    layout.addWidget(title)
    
    # VFS bilgi paneli
    self.vfs_info_panel = QWidget()
    self.setup_vfs_info_panel()
    layout.addWidget(self.vfs_info_panel)
    
    # Ana panel (yan yana)
    main_panel = QWidget()
    main_layout = QHBoxLayout(main_panel)
    main_layout.setSpacing(12)
    
    # Sol panel - Dizin ağacı
    left_panel = QWidget()
    left_layout = QVBoxLayout(left_panel)
    left_layout.setContentsMargins(0, 0, 0, 0)
    
    # Dizin ağacı
    tree_label = QLabel("🗂️ Dizinler")
    tree_label.setStyleSheet("font-weight: bold; margin-bottom: 5px;")
    left_layout.addWidget(tree_label)
    
    self.tree_view = QTreeView()
    self.tree_view.setHeaderHidden(True)
    self.tree_view.setMinimumWidth(200)
    left_layout.addWidget(self.tree_view)
    
    # Sağ panel - Dosya listesi
    right_panel = QWidget()
    right_layout = QVBoxLayout(right_panel)
    right_layout.setContentsMargins(0, 0, 0, 0)
    
    # Dosya listesi başlığı
    files_label = QLabel("📄 Dosyalar")
    files_label.setStyleSheet("font-weight: bold; margin-bottom: 5px;")
    right_layout.addWidget(files_label)
    
    # Dosya listesi
    self.file_list = QListWidget()
    
    # Multiple selection mode için ayar
    if self.mode == FilePickerMode.MULTIPLE_FILES:
        self.file_list.setSelectionMode(QAbstractItemView.SelectionMode.MultiSelection)
        self.file_list.setStyleSheet("""
            QListWidget::item:selected {
                background-color: #e3f2fd;
                color: #1976d2;
                border: 2px solid #2196f3;
            }
        """)
    else:
        self.file_list.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
    
    self.file_list.setMinimumHeight(300)
    self.file_list.itemDoubleClicked.connect(self.file_double_clicked)
    self.file_list.itemSelectionChanged.connect(self.selection_changed)
    right_layout.addWidget(self.file_list)
    
    # Dosya adı girişi (save mode için)
    if self.mode == FilePickerMode.SAVE_FILE:
        filename_label = QLabel("📝 Dosya Adı:")
        right_layout.addWidget(filename_label)
        
        self.filename_input = QLineEdit()
        self.filename_input.setPlaceholderText("dosya_adi.txt")
        self.filename_input.textChanged.connect(self.validate_selection)
        right_layout.addWidget(self.filename_input)
    
    # Multiple files için seçim bilgisi
    if self.mode == FilePickerMode.MULTIPLE_FILES:
        self.selection_info = QLabel("0 dosya seçildi")
        self.selection_info.setStyleSheet("color: #666; font-style: italic;")
        right_layout.addWidget(self.selection_info)
    
    # Panel boyutları
    left_panel.setMaximumWidth(250)
    
    main_layout.addWidget(left_panel)
    main_layout.addWidget(right_panel)
    layout.addWidget(main_panel)
    
    # Filtre seçimi
    if self.filters:
        filter_label = QLabel("🔍 Dosya Türü:")
        layout.addWidget(filter_label)
        
        self.filter_combo = QComboBox()
        for filter_item in self.filters:
            self.filter_combo.addItem(filter_item.description, filter_item.extensions)
        self.filter_combo.currentTextChanged.connect(self.filter_changed)
        layout.addWidget(self.filter_combo)
    
    # Butonlar
    button_layout = QHBoxLayout()
    
    # Cancel butonu
    self.cancel_button = QPushButton("❌ İptal")
    self.cancel_button.clicked.connect(self.reject)
    button_layout.addWidget(self.cancel_button)
    
    button_layout.addStretch()
    
    # Ana eylem butonu
    action_text = {
        FilePickerMode.OPEN_FILE: "📂 Aç",
        FilePickerMode.SAVE_FILE: "💾 Kaydet",
        FilePickerMode.SELECT_DIRECTORY: "📂 Seç",
        FilePickerMode.MULTIPLE_FILES: "📋 Seç"
    }
    
    self.action_button = QPushButton(action_text.get(self.mode, "✅ Tamam"))
    self.action_button.setEnabled(False)
    self.action_button.clicked.connect(self.accept_selection)
    button_layout.addWidget(self.action_button)
    
    layout.addLayout(button_layout)
    self.setLayout(layout)

# Export yapılacak fonksiyonlar
__all__ = [
    'FilePickerMode', 
    'FilePickerFilter', 
    'CloudFilePicker',
    'open_file_dialog', 
    'save_file_dialog', 
    'select_directory_dialog',
    'select_multiple_files_dialog',
    'show_file_picker'
] 



__all__ = ['BaseResolver', 'Resolver']

from .error import *
from .nodes import *

import re

class ResolverError(YAMLError):
    pass

class BaseResolver:

    DEFAULT_SCALAR_TAG = 'tag:yaml.org,2002:str'
    DEFAULT_SEQUENCE_TAG = 'tag:yaml.org,2002:seq'
    DEFAULT_MAPPING_TAG = 'tag:yaml.org,2002:map'

    yaml_implicit_resolvers = {}
    yaml_path_resolvers = {}

    def __init__(self):
        self.resolver_exact_paths = []
        self.resolver_prefix_paths = []

    @classmethod
    def add_implicit_resolver(cls, tag, regexp, first):
        if not 'yaml_implicit_resolvers' in cls.__dict__:
            implicit_resolvers = {}
            for key in cls.yaml_implicit_resolvers:
                implicit_resolvers[key] = cls.yaml_implicit_resolvers[key][:]
            cls.yaml_implicit_resolvers = implicit_resolvers
        if first is None:
            first = [None]
        for ch in first:
            cls.yaml_implicit_resolvers.setdefault(ch, []).append((tag, regexp))

    @classmethod
    def add_path_resolver(cls, tag, path, kind=None):
        # Note: `add_path_resolver` is experimental.  The API could be changed.
        # `new_path` is a pattern that is matched against the path from the
        # root to the node that is being considered.  `node_path` elements are
        # tuples `(node_check, index_check)`.  `node_check` is a node class:
        # `ScalarNode`, `SequenceNode`, `MappingNode` or `None`.  `None`
        # matches any kind of a node.  `index_check` could be `None`, a boolean
        # value, a string value, or a number.  `None` and `False` match against
        # any _value_ of sequence and mapping nodes.  `True` matches against
        # any _key_ of a mapping node.  A string `index_check` matches against
        # a mapping value that corresponds to a scalar key which content is
        # equal to the `index_check` value.  An integer `index_check` matches
        # against a sequence value with the index equal to `index_check`.
        if not 'yaml_path_resolvers' in cls.__dict__:
            cls.yaml_path_resolvers = cls.yaml_path_resolvers.copy()
        new_path = []
        for element in path:
            if isinstance(element, (list, tuple)):
                if len(element) == 2:
                    node_check, index_check = element
                elif len(element) == 1:
                    node_check = element[0]
                    index_check = True
                else:
                    raise ResolverError("Invalid path element: %s" % element)
            else:
                node_check = None
                index_check = element
            if node_check is str:
                node_check = ScalarNode
            elif node_check is list:
                node_check = SequenceNode
            elif node_check is dict:
                node_check = MappingNode
            elif node_check not in [ScalarNode, SequenceNode, MappingNode]  \
                    and not isinstance(node_check, str) \
                    and node_check is not None:
                raise ResolverError("Invalid node checker: %s" % node_check)
            if not isinstance(index_check, (str, int))  \
                    and index_check is not None:
                raise ResolverError("Invalid index checker: %s" % index_check)
            new_path.append((node_check, index_check))
        if kind is str:
            kind = ScalarNode
        elif kind is list:
            kind = SequenceNode
        elif kind is dict:
            kind = MappingNode
        elif kind not in [ScalarNode, SequenceNode, MappingNode]    \
                and kind is not None:
            raise ResolverError("Invalid node kind: %s" % kind)
        cls.yaml_path_resolvers[tuple(new_path), kind] = tag

    def descend_resolver(self, current_node, current_index):
        if not self.yaml_path_resolvers:
            return
        exact_paths = {}
        prefix_paths = []
        if current_node:
            depth = len(self.resolver_prefix_paths)
            for path, kind in self.resolver_prefix_paths[-1]:
                if self.check_resolver_prefix(depth, path, kind,
                        current_node, current_index):
                    if len(path) > depth:
                        prefix_paths.append((path, kind))
                    else:
                        exact_paths[kind] = self.yaml_path_resolvers[path, kind]
        else:
            for path, kind in self.yaml_path_resolvers:
                if not path:
                    exact_paths[kind] = self.yaml_path_resolvers[path, kind]
                else:
                    prefix_paths.append((path, kind))
        self.resolver_exact_paths.append(exact_paths)
        self.resolver_prefix_paths.append(prefix_paths)

    def ascend_resolver(self):
        if not self.yaml_path_resolvers:
            return
        self.resolver_exact_paths.pop()
        self.resolver_prefix_paths.pop()

    def check_resolver_prefix(self, depth, path, kind,
            current_node, current_index):
        node_check, index_check = path[depth-1]
        if isinstance(node_check, str):
            if current_node.tag != node_check:
                return
        elif node_check is not None:
            if not isinstance(current_node, node_check):
                return
        if index_check is True and current_index is not None:
            return
        if (index_check is False or index_check is None)    \
                and current_index is None:
            return
        if isinstance(index_check, str):
            if not (isinstance(current_index, ScalarNode)
                    and index_check == current_index.value):
                return
        elif isinstance(index_check, int) and not isinstance(index_check, bool):
            if index_check != current_index:
                return
        return True

    def resolve(self, kind, value, implicit):
        if kind is ScalarNode and implicit[0]:
            if value == '':
                resolvers = self.yaml_implicit_resolvers.get('', [])
            else:
                resolvers = self.yaml_implicit_resolvers.get(value[0], [])
            wildcard_resolvers = self.yaml_implicit_resolvers.get(None, [])
            for tag, regexp in resolvers + wildcard_resolvers:
                if regexp.match(value):
                    return tag
            implicit = implicit[1]
        if self.yaml_path_resolvers:
            exact_paths = self.resolver_exact_paths[-1]
            if kind in exact_paths:
                return exact_paths[kind]
            if None in exact_paths:
                return exact_paths[None]
        if kind is ScalarNode:
            return self.DEFAULT_SCALAR_TAG
        elif kind is SequenceNode:
            return self.DEFAULT_SEQUENCE_TAG
        elif kind is MappingNode:
            return self.DEFAULT_MAPPING_TAG

class Resolver(BaseResolver):
    pass

Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:bool',
        re.compile(r'''^(?:yes|Yes|YES|no|No|NO
                    |true|True|TRUE|false|False|FALSE
                    |on|On|ON|off|Off|OFF)$''', re.X),
        list('yYnNtTfFoO'))

Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:float',
        re.compile(r'''^(?:[-+]?(?:[0-9][0-9_]*)\.[0-9_]*(?:[eE][-+][0-9]+)?
                    |\.[0-9][0-9_]*(?:[eE][-+][0-9]+)?
                    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*
                    |[-+]?\.(?:inf|Inf|INF)
                    |\.(?:nan|NaN|NAN))$''', re.X),
        list('-+0123456789.'))

Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:int',
        re.compile(r'''^(?:[-+]?0b[0-1_]+
                    |[-+]?0[0-7_]+
                    |[-+]?(?:0|[1-9][0-9_]*)
                    |[-+]?0x[0-9a-fA-F_]+
                    |[-+]?[1-9][0-9_]*(?::[0-5]?[0-9])+)$''', re.X),
        list('-+0123456789'))

Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:merge',
        re.compile(r'^(?:<<)$'),
        ['<'])

Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:null',
        re.compile(r'''^(?: ~
                    |null|Null|NULL
                    | )$''', re.X),
        ['~', 'n', 'N', ''])

Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:timestamp',
        re.compile(r'''^(?:[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]
                    |[0-9][0-9][0-9][0-9] -[0-9][0-9]? -[0-9][0-9]?
                     (?:[Tt]|[ \t]+)[0-9][0-9]?
                     :[0-9][0-9] :[0-9][0-9] (?:\.[0-9]*)?
                     (?:[ \t]*(?:Z|[-+][0-9][0-9]?(?::[0-9][0-9])?))?)$''', re.X),
        list('0123456789'))

Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:value',
        re.compile(r'^(?:=)$'),
        ['='])

# The following resolver is only for documentation purposes. It cannot work
# because plain scalars cannot start with '!', '&', or '*'.
Resolver.add_implicit_resolver(
        'tag:yaml.org,2002:yaml',
        re.compile(r'^(?:!|&|\*)$'),
        list('!&*'))




# mypy: allow-untyped-defs
"""Discover and run std-library "unittest" style tests."""

from __future__ import annotations

from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Iterator
from enum import auto
from enum import Enum
import inspect
import sys
import traceback
import types
from typing import TYPE_CHECKING
from typing import Union

import _pytest._code
from _pytest.compat import is_async_function
from _pytest.config import hookimpl
from _pytest.fixtures import FixtureRequest
from _pytest.monkeypatch import MonkeyPatch
from _pytest.nodes import Collector
from _pytest.nodes import Item
from _pytest.outcomes import exit
from _pytest.outcomes import fail
from _pytest.outcomes import skip
from _pytest.outcomes import xfail
from _pytest.python import Class
from _pytest.python import Function
from _pytest.python import Module
from _pytest.runner import CallInfo
import pytest


if sys.version_info[:2] < (3, 11):
    from exceptiongroup import ExceptionGroup

if TYPE_CHECKING:
    import unittest

    import twisted.trial.unittest


_SysExcInfoType = Union[
    tuple[type[BaseException], BaseException, types.TracebackType],
    tuple[None, None, None],
]


def pytest_pycollect_makeitem(
    collector: Module | Class, name: str, obj: object
) -> UnitTestCase | None:
    try:
        # Has unittest been imported?
        ut = sys.modules["unittest"]
        # Is obj a subclass of unittest.TestCase?
        # Type ignored because `ut` is an opaque module.
        if not issubclass(obj, ut.TestCase):  # type: ignore
            return None
    except Exception:
        return None
    # Is obj a concrete class?
    # Abstract classes can't be instantiated so no point collecting them.
    if inspect.isabstract(obj):
        return None
    # Yes, so let's collect it.
    return UnitTestCase.from_parent(collector, name=name, obj=obj)


class UnitTestCase(Class):
    # Marker for fixturemanger.getfixtureinfo()
    # to declare that our children do not support funcargs.
    nofuncargs = True

    def newinstance(self):
        # TestCase __init__ takes the method (test) name. The TestCase
        # constructor treats the name "runTest" as a special no-op, so it can be
        # used when a dummy instance is needed. While unittest.TestCase has a
        # default, some subclasses omit the default (#9610), so always supply
        # it.
        return self.obj("runTest")

    def collect(self) -> Iterable[Item | Collector]:
        from unittest import TestLoader

        cls = self.obj
        if not getattr(cls, "__test__", True):
            return

        skipped = _is_skipped(cls)
        if not skipped:
            self._register_unittest_setup_method_fixture(cls)
            self._register_unittest_setup_class_fixture(cls)
            self._register_setup_class_fixture()

        self.session._fixturemanager.parsefactories(self.newinstance(), self.nodeid)

        loader = TestLoader()
        foundsomething = False
        for name in loader.getTestCaseNames(self.obj):
            x = getattr(self.obj, name)
            if not getattr(x, "__test__", True):
                continue
            yield TestCaseFunction.from_parent(self, name=name)
            foundsomething = True

        if not foundsomething:
            runtest = getattr(self.obj, "runTest", None)
            if runtest is not None:
                ut = sys.modules.get("twisted.trial.unittest", None)
                if ut is None or runtest != ut.TestCase.runTest:
                    yield TestCaseFunction.from_parent(self, name="runTest")

    def _register_unittest_setup_class_fixture(self, cls: type) -> None:
        """Register an auto-use fixture to invoke setUpClass and
        tearDownClass (#517)."""
        setup = getattr(cls, "setUpClass", None)
        teardown = getattr(cls, "tearDownClass", None)
        if setup is None and teardown is None:
            return None
        cleanup = getattr(cls, "doClassCleanups", lambda: None)

        def process_teardown_exceptions() -> None:
            # tearDown_exceptions is a list set in the class containing exc_infos for errors during
            # teardown for the class.
            exc_infos = getattr(cls, "tearDown_exceptions", None)
            if not exc_infos:
                return
            exceptions = [exc for (_, exc, _) in exc_infos]
            # If a single exception, raise it directly as this provides a more readable
            # error (hopefully this will improve in #12255).
            if len(exceptions) == 1:
                raise exceptions[0]
            else:
                raise ExceptionGroup("Unittest class cleanup errors", exceptions)

        def unittest_setup_class_fixture(
            request: FixtureRequest,
        ) -> Generator[None]:
            cls = request.cls
            if _is_skipped(cls):
                reason = cls.__unittest_skip_why__
                raise pytest.skip.Exception(reason, _use_item_location=True)
            if setup is not None:
                try:
                    setup()
                # unittest does not call the cleanup function for every BaseException, so we
                # follow this here.
                except Exception:
                    cleanup()
                    process_teardown_exceptions()
                    raise
            yield
            try:
                if teardown is not None:
                    teardown()
            finally:
                cleanup()
                process_teardown_exceptions()

        self.session._fixturemanager._register_fixture(
            # Use a unique name to speed up lookup.
            name=f"_unittest_setUpClass_fixture_{cls.__qualname__}",
            func=unittest_setup_class_fixture,
            nodeid=self.nodeid,
            scope="class",
            autouse=True,
        )

    def _register_unittest_setup_method_fixture(self, cls: type) -> None:
        """Register an auto-use fixture to invoke setup_method and
        teardown_method (#517)."""
        setup = getattr(cls, "setup_method", None)
        teardown = getattr(cls, "teardown_method", None)
        if setup is None and teardown is None:
            return None

        def unittest_setup_method_fixture(
            request: FixtureRequest,
        ) -> Generator[None]:
            self = request.instance
            if _is_skipped(self):
                reason = self.__unittest_skip_why__
                raise pytest.skip.Exception(reason, _use_item_location=True)
            if setup is not None:
                setup(self, request.function)
            yield
            if teardown is not None:
                teardown(self, request.function)

        self.session._fixturemanager._register_fixture(
            # Use a unique name to speed up lookup.
            name=f"_unittest_setup_method_fixture_{cls.__qualname__}",
            func=unittest_setup_method_fixture,
            nodeid=self.nodeid,
            scope="function",
            autouse=True,
        )


class TestCaseFunction(Function):
    nofuncargs = True
    _excinfo: list[_pytest._code.ExceptionInfo[BaseException]] | None = None

    def _getinstance(self):
        assert isinstance(self.parent, UnitTestCase)
        return self.parent.obj(self.name)

    # Backward compat for pytest-django; can be removed after pytest-django
    # updates + some slack.
    @property
    def _testcase(self):
        return self.instance

    def setup(self) -> None:
        # A bound method to be called during teardown() if set (see 'runtest()').
        self._explicit_tearDown: Callable[[], None] | None = None
        super().setup()

    def teardown(self) -> None:
        if self._explicit_tearDown is not None:
            self._explicit_tearDown()
            self._explicit_tearDown = None
        self._obj = None
        del self._instance
        super().teardown()

    def startTest(self, testcase: unittest.TestCase) -> None:
        pass

    def _addexcinfo(self, rawexcinfo: _SysExcInfoType) -> None:
        rawexcinfo = _handle_twisted_exc_info(rawexcinfo)
        try:
            excinfo = _pytest._code.ExceptionInfo[BaseException].from_exc_info(
                rawexcinfo  # type: ignore[arg-type]
            )
            # Invoke the attributes to trigger storing the traceback
            # trial causes some issue there.
            _ = excinfo.value
            _ = excinfo.traceback
        except TypeError:
            try:
                try:
                    values = traceback.format_exception(*rawexcinfo)
                    values.insert(
                        0,
                        "NOTE: Incompatible Exception Representation, "
                        "displaying natively:\n\n",
                    )
                    fail("".join(values), pytrace=False)
                except (fail.Exception, KeyboardInterrupt):
                    raise
                except BaseException:
                    fail(
                        "ERROR: Unknown Incompatible Exception "
                        f"representation:\n{rawexcinfo!r}",
                        pytrace=False,
                    )
            except KeyboardInterrupt:
                raise
            except fail.Exception:
                excinfo = _pytest._code.ExceptionInfo.from_current()
        self.__dict__.setdefault("_excinfo", []).append(excinfo)

    def addError(
        self, testcase: unittest.TestCase, rawexcinfo: _SysExcInfoType
    ) -> None:
        try:
            if isinstance(rawexcinfo[1], exit.Exception):
                exit(rawexcinfo[1].msg)
        except TypeError:
            pass
        self._addexcinfo(rawexcinfo)

    def addFailure(
        self, testcase: unittest.TestCase, rawexcinfo: _SysExcInfoType
    ) -> None:
        self._addexcinfo(rawexcinfo)

    def addSkip(self, testcase: unittest.TestCase, reason: str) -> None:
        try:
            raise pytest.skip.Exception(reason, _use_item_location=True)
        except skip.Exception:
            self._addexcinfo(sys.exc_info())

    def addExpectedFailure(
        self,
        testcase: unittest.TestCase,
        rawexcinfo: _SysExcInfoType,
        reason: str = "",
    ) -> None:
        try:
            xfail(str(reason))
        except xfail.Exception:
            self._addexcinfo(sys.exc_info())

    def addUnexpectedSuccess(
        self,
        testcase: unittest.TestCase,
        reason: twisted.trial.unittest.Todo | None = None,
    ) -> None:
        msg = "Unexpected success"
        if reason:
            msg += f": {reason.reason}"
        # Preserve unittest behaviour - fail the test. Explicitly not an XPASS.
        try:
            fail(msg, pytrace=False)
        except fail.Exception:
            self._addexcinfo(sys.exc_info())

    def addSuccess(self, testcase: unittest.TestCase) -> None:
        pass

    def stopTest(self, testcase: unittest.TestCase) -> None:
        pass

    def addDuration(self, testcase: unittest.TestCase, elapsed: float) -> None:
        pass

    def runtest(self) -> None:
        from _pytest.debugging import maybe_wrap_pytest_function_for_tracing

        testcase = self.instance
        assert testcase is not None

        maybe_wrap_pytest_function_for_tracing(self)

        # Let the unittest framework handle async functions.
        if is_async_function(self.obj):
            testcase(result=self)
        else:
            # When --pdb is given, we want to postpone calling tearDown() otherwise
            # when entering the pdb prompt, tearDown() would have probably cleaned up
            # instance variables, which makes it difficult to debug.
            # Arguably we could always postpone tearDown(), but this changes the moment where the
            # TestCase instance interacts with the results object, so better to only do it
            # when absolutely needed.
            # We need to consider if the test itself is skipped, or the whole class.
            assert isinstance(self.parent, UnitTestCase)
            skipped = _is_skipped(self.obj) or _is_skipped(self.parent.obj)
            if self.config.getoption("usepdb") and not skipped:
                self._explicit_tearDown = testcase.tearDown
                setattr(testcase, "tearDown", lambda *args: None)

            # We need to update the actual bound method with self.obj, because
            # wrap_pytest_function_for_tracing replaces self.obj by a wrapper.
            setattr(testcase, self.name, self.obj)
            try:
                testcase(result=self)
            finally:
                delattr(testcase, self.name)

    def _traceback_filter(
        self, excinfo: _pytest._code.ExceptionInfo[BaseException]
    ) -> _pytest._code.Traceback:
        traceback = super()._traceback_filter(excinfo)
        ntraceback = traceback.filter(
            lambda x: not x.frame.f_globals.get("__unittest"),
        )
        if not ntraceback:
            ntraceback = traceback
        return ntraceback


@hookimpl(tryfirst=True)
def pytest_runtest_makereport(item: Item, call: CallInfo[None]) -> None:
    if isinstance(item, TestCaseFunction):
        if item._excinfo:
            call.excinfo = item._excinfo.pop(0)
            try:
                del call.result
            except AttributeError:
                pass

    # Convert unittest.SkipTest to pytest.skip.
    # This is actually only needed for nose, which reuses unittest.SkipTest for
    # its own nose.SkipTest. For unittest TestCases, SkipTest is already
    # handled internally, and doesn't reach here.
    unittest = sys.modules.get("unittest")
    if unittest and call.excinfo and isinstance(call.excinfo.value, unittest.SkipTest):
        excinfo = call.excinfo
        call2 = CallInfo[None].from_call(
            lambda: pytest.skip(str(excinfo.value)), call.when
        )
        call.excinfo = call2.excinfo


def _is_skipped(obj) -> bool:
    """Return True if the given object has been marked with @unittest.skip."""
    return bool(getattr(obj, "__unittest_skip__", False))


def pytest_configure() -> None:
    """Register the TestCaseFunction class as an IReporter if twisted.trial is available."""
    if _get_twisted_version() is not TwistedVersion.NotInstalled:
        from twisted.trial.itrial import IReporter
        from zope.interface import classImplements

        classImplements(TestCaseFunction, IReporter)


class TwistedVersion(Enum):
    """
    The Twisted version installed in the environment.

    We have different workarounds in place for different versions of Twisted.
    """

    # Twisted version 24 or prior.
    Version24 = auto()
    # Twisted version 25 or later.
    Version25 = auto()
    # Twisted version is not available.
    NotInstalled = auto()


def _get_twisted_version() -> TwistedVersion:
    # We need to check if "twisted.trial.unittest" is specifically present in sys.modules.
    # This is because we intend to integrate with Trial only when it's actively running
    # the test suite, but not needed when only other Twisted components are in use.
    if "twisted.trial.unittest" not in sys.modules:
        return TwistedVersion.NotInstalled

    import importlib.metadata

    import packaging.version

    version_str = importlib.metadata.version("twisted")
    version = packaging.version.parse(version_str)
    if version.major <= 24:
        return TwistedVersion.Version24
    else:
        return TwistedVersion.Version25


# Name of the attribute in `twisted.python.Failure` instances that stores
# the `sys.exc_info()` tuple.
# See twisted.trial support in `pytest_runtest_protocol`.
TWISTED_RAW_EXCINFO_ATTR = "_twisted_raw_excinfo"


@hookimpl(wrapper=True)
def pytest_runtest_protocol(item: Item) -> Iterator[None]:
    if _get_twisted_version() is TwistedVersion.Version24:
        import twisted.python.failure as ut

        # Monkeypatch `Failure.__init__` to store the raw exception info.
        original__init__ = ut.Failure.__init__

        def store_raw_exception_info(
            self, exc_value=None, exc_type=None, exc_tb=None, captureVars=None
        ):  # pragma: no cover
            if exc_value is None:
                raw_exc_info = sys.exc_info()
            else:
                if exc_type is None:
                    exc_type = type(exc_value)
                if exc_tb is None:
                    exc_tb = sys.exc_info()[2]
                raw_exc_info = (exc_type, exc_value, exc_tb)
            setattr(self, TWISTED_RAW_EXCINFO_ATTR, tuple(raw_exc_info))
            try:
                original__init__(
                    self, exc_value, exc_type, exc_tb, captureVars=captureVars
                )
            except TypeError:  # pragma: no cover
                original__init__(self, exc_value, exc_type, exc_tb)

        with MonkeyPatch.context() as patcher:
            patcher.setattr(ut.Failure, "__init__", store_raw_exception_info)
            return (yield)
    else:
        return (yield)


def _handle_twisted_exc_info(
    rawexcinfo: _SysExcInfoType | BaseException,
) -> _SysExcInfoType:
    """
    Twisted passes a custom Failure instance to `addError()` instead of using `sys.exc_info()`.
    Therefore, if `rawexcinfo` is a `Failure` instance, convert it into the equivalent `sys.exc_info()` tuple
    as expected by pytest.
    """
    twisted_version = _get_twisted_version()
    if twisted_version is TwistedVersion.NotInstalled:
        # Unfortunately, because we cannot import `twisted.python.failure` at the top of the file
        # and use it in the signature, we need to use `type:ignore` here because we cannot narrow
        # the type properly in the `if` statement above.
        return rawexcinfo  # type:ignore[return-value]
    elif twisted_version is TwistedVersion.Version24:
        # Twisted calls addError() passing its own classes (like `twisted.python.Failure`), which violates
        # the `addError()` signature, so we extract the original `sys.exc_info()` tuple which is stored
        # in the object.
        if hasattr(rawexcinfo, TWISTED_RAW_EXCINFO_ATTR):
            saved_exc_info = getattr(rawexcinfo, TWISTED_RAW_EXCINFO_ATTR)
            # Delete the attribute from the original object to avoid leaks.
            delattr(rawexcinfo, TWISTED_RAW_EXCINFO_ATTR)
            return saved_exc_info  # type:ignore[no-any-return]
        return rawexcinfo  # type:ignore[return-value]
    elif twisted_version is TwistedVersion.Version25:
        if isinstance(rawexcinfo, BaseException):
            import twisted.python.failure

            if isinstance(rawexcinfo, twisted.python.failure.Failure):
                tb = rawexcinfo.__traceback__
                if tb is None:
                    tb = sys.exc_info()[2]
                return type(rawexcinfo.value), rawexcinfo.value, tb

        return rawexcinfo  # type:ignore[return-value]
    else:
        # Ideally we would use assert_never() here, but it is not available in all Python versions
        # we support, plus we do not require `type_extensions` currently.
        assert False, f"Unexpected Twisted version: {twisted_version}"



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

import abc
import typing

from cryptography.hazmat.bindings._rust import openssl as rust_openssl
from cryptography.hazmat.primitives import _serialization, hashes
from cryptography.hazmat.primitives.asymmetric import utils as asym_utils
from cryptography.utils import Buffer


class DSAParameters(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def generate_private_key(self) -> DSAPrivateKey:
        """
        Generates and returns a DSAPrivateKey.
        """

    @abc.abstractmethod
    def parameter_numbers(self) -> DSAParameterNumbers:
        """
        Returns a DSAParameterNumbers.
        """


DSAParametersWithNumbers = DSAParameters
DSAParameters.register(rust_openssl.dsa.DSAParameters)


class DSAPrivateKey(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def key_size(self) -> int:
        """
        The bit length of the prime modulus.
        """

    @abc.abstractmethod
    def public_key(self) -> DSAPublicKey:
        """
        The DSAPublicKey associated with this private key.
        """

    @abc.abstractmethod
    def parameters(self) -> DSAParameters:
        """
        The DSAParameters object associated with this private key.
        """

    @abc.abstractmethod
    def sign(
        self,
        data: Buffer,
        algorithm: asym_utils.Prehashed | hashes.HashAlgorithm,
    ) -> bytes:
        """
        Signs the data
        """

    @abc.abstractmethod
    def private_numbers(self) -> DSAPrivateNumbers:
        """
        Returns a DSAPrivateNumbers.
        """

    @abc.abstractmethod
    def private_bytes(
        self,
        encoding: _serialization.Encoding,
        format: _serialization.PrivateFormat,
        encryption_algorithm: _serialization.KeySerializationEncryption,
    ) -> bytes:
        """
        Returns the key serialized as bytes.
        """

    @abc.abstractmethod
    def __copy__(self) -> DSAPrivateKey:
        """
        Returns a copy.
        """


DSAPrivateKeyWithSerialization = DSAPrivateKey
DSAPrivateKey.register(rust_openssl.dsa.DSAPrivateKey)


class DSAPublicKey(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def key_size(self) -> int:
        """
        The bit length of the prime modulus.
        """

    @abc.abstractmethod
    def parameters(self) -> DSAParameters:
        """
        The DSAParameters object associated with this public key.
        """

    @abc.abstractmethod
    def public_numbers(self) -> DSAPublicNumbers:
        """
        Returns a DSAPublicNumbers.
        """

    @abc.abstractmethod
    def public_bytes(
        self,
        encoding: _serialization.Encoding,
        format: _serialization.PublicFormat,
    ) -> bytes:
        """
        Returns the key serialized as bytes.
        """

    @abc.abstractmethod
    def verify(
        self,
        signature: Buffer,
        data: Buffer,
        algorithm: asym_utils.Prehashed | hashes.HashAlgorithm,
    ) -> None:
        """
        Verifies the signature of the data.
        """

    @abc.abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        Checks equality.
        """

    @abc.abstractmethod
    def __copy__(self) -> DSAPublicKey:
        """
        Returns a copy.
        """


DSAPublicKeyWithSerialization = DSAPublicKey
DSAPublicKey.register(rust_openssl.dsa.DSAPublicKey)

DSAPrivateNumbers = rust_openssl.dsa.DSAPrivateNumbers
DSAPublicNumbers = rust_openssl.dsa.DSAPublicNumbers
DSAParameterNumbers = rust_openssl.dsa.DSAParameterNumbers


def generate_parameters(
    key_size: int, backend: typing.Any = None
) -> DSAParameters:
    if key_size not in (1024, 2048, 3072, 4096):
        raise ValueError("Key size must be 1024, 2048, 3072, or 4096 bits.")

    return rust_openssl.dsa.generate_parameters(key_size)


def generate_private_key(
    key_size: int, backend: typing.Any = None
) -> DSAPrivateKey:
    parameters = generate_parameters(key_size)
    return parameters.generate_private_key()
